# -*- coding: utf-8 -*-
'''
Managing Images in OpenStack Glance
===================================
'''
# Import python libs
from __future__ import absolute_import, print_function, unicode_literals
import logging
import time

# Import salt libs

# Import OpenStack libs
try:
    from keystoneclient.exceptions import \
        Unauthorized as kstone_Unauthorized
    HAS_KEYSTONE = True
except ImportError:
    try:
        from keystoneclient.apiclient.exceptions import \
            Unauthorized as kstone_Unauthorized
        HAS_KEYSTONE = True
    except ImportError:
        HAS_KEYSTONE = False

try:
    from glanceclient.exc import \
        HTTPUnauthorized as glance_Unauthorized
    HAS_GLANCE = True
except ImportError:
    HAS_GLANCE = False

log = logging.getLogger(__name__)


def __virtual__():
    '''
    Only load if dependencies are loaded
    '''
    return HAS_KEYSTONE and HAS_GLANCE


def _find_image(name):
    '''
    Tries to find image with given name, returns
        - image, 'Found image <name>'
        - None, 'No such image found'
        - False, 'Found more than one image with given name'
    '''
    try:
        images = __salt__['glance.image_list'](name=name)
    except kstone_Unauthorized:
        return False, 'keystoneclient: Unauthorized'
    except glance_Unauthorized:
        return False, 'glanceclient: Unauthorized'
    log.debug('Got images: {0}'.format(images))

    if type(images) is dict and len(images) == 1 and 'images' in images:
        images = images['images']

    images_list = images.values() if type(images) is dict else images

    if len(images_list) == 0:
        return None, 'No image with name "{0}"'.format(name)
    elif len(images_list) == 1:
        return images_list[0], 'Found image {0}'.format(name)
    elif len(images_list) > 1:
        return False, 'Found more than one image with given name'
    else:
        raise NotImplementedError


def image_present(name, visibility='public', protected=None,
        checksum=None, location=None, disk_format='raw', wait_for=None,
        timeout=30):
    '''
    Checks if given image is present with properties
    set as specified.

    An image should got through the stages 'queued', 'saving'
    before becoming 'active'. The attribute 'checksum' can
    only be checked once the image is active.
    If you don't specify 'wait_for' but 'checksum' the function
    will wait for the image to become active before comparing
    checksums. If you don't specify checksum either the function
    will return when the image reached 'saving'.
    The default timeout for both is 30 seconds.

    Supported properties:
      - visibility ('public' or 'private')
      - protected (bool)
      - checksum (string, md5sum)
      - location (URL, to copy from)
      - disk_format ('raw' (default), 'vhd', 'vhdx', 'vmdk', 'vdi', 'iso',
        'qcow2', 'aki', 'ari' or 'ami')
    '''
    ret = {'name': name,
            'changes': {},
            'result': True,
            'comment': '',
            }
    acceptable = ['queued', 'saving', 'active']
    if wait_for is None and checksum is None:
        wait_for = 'saving'
    elif wait_for is None and checksum is not None:
        wait_for = 'active'

    # Just pop states until we reach the
    # first acceptable one:
    while len(acceptable) > 1:
        if acceptable[0] == wait_for:
            break
        else:
            acceptable.pop(0)

    image, msg = _find_image(name)
    if image is False:
        if __opts__['test']:
            ret['result'] = None
        else:
            ret['result'] = False
        ret['comment'] = msg
        return ret
    log.debug(msg)
    # No image yet and we know where to get one
    if image is None and location is not None:
        if __opts__['test']:
            ret['result'] = None
            ret['comment'] = 'glance.image_present would ' \
                'create an image from {0}'.format(location)
            return ret
        image = __salt__['glance.image_create'](name=name,
            protected=protected, visibility=visibility,
            location=location, disk_format=disk_format)
        log.debug('Created new image:\n{0}'.format(image))
        ret['changes'] = {
            name:
                {
                    'new':
                        {
                        'id': image['id']
                        },
                    'old': None
                }
            }
        timer = timeout
        # Kinda busy-loopy but I don't think the Glance
        # API has events we can listen for
        while timer > 0:
            if 'status' in image and \
                    image['status'] in acceptable:
                log.debug('Image {0} has reached status {1}'.format(
                    image['name'], image['status']))
                break
            else:
                timer -= 5
                time.sleep(5)
                image, msg = _find_image(name)
                if not image:
                    ret['result'] = False
                    ret['comment'] += 'Created image {0} '.format(
                        name) + ' vanished:\n' + msg
                    return ret
        if timer <= 0 and image['status'] not in acceptable:
            ret['result'] = False
            ret['comment'] += 'Image didn\'t reach an acceptable '+\
                    'state ({0}) before timeout:\n'.format(acceptable)+\
                    '\tLast status was "{0}".\n'.format(image['status'])

    # There's no image but where would I get one??
    elif location is None:
        if __opts__['test']:
            ret['result'] = None
            ret['comment'] = 'No location to copy image from specified,\n' +\
                         'glance.image_present would not create one'
        else:
            ret['result'] = False
            ret['comment'] = 'No location to copy image from specified,\n' +\
                         'not creating a new image.'
        return ret

    # If we've created a new image also return its last status:
    if name in ret['changes']:
        ret['changes'][name]['new']['status'] = image['status']

    if visibility:
        if image['visibility'] != visibility:
            old_value = image['visibility']
            if not __opts__['test']:
                image = __salt__['glance.image_update'](
                    id=image['id'], visibility=visibility)
            # Check if image_update() worked:
            if image['visibility'] != visibility:
                if not __opts__['test']:
                    ret['result'] = False
                elif __opts__['test']:
                    ret['result'] = None
                ret['comment'] += '"visibility" is {0}, '\
                    'should be {1}.\n'.format(image['visibility'],
                        visibility)
            else:
                if 'new' in ret['changes']:
                    ret['changes']['new']['visibility'] = visibility
                else:
                    ret['changes']['new'] = {'visibility': visibility}
                if 'old' in ret['changes']:
                    ret['changes']['old']['visibility'] = old_value
                else:
                    ret['changes']['old'] = {'visibility': old_value}
        else:
            ret['comment'] += '"visibility" is correct ({0}).\n'.format(
                visibility)
    if protected is not None:
        if not isinstance(protected, bool) or image['protected'] ^ protected:
            if not __opts__['test']:
                ret['result'] = False
            else:
                ret['result'] = None
            ret['comment'] += '"protected" is {0}, should be {1}.\n'.format(
                image['protected'], protected)
        else:
            ret['comment'] += '"protected" is correct ({0}).\n'.format(
                protected)
    if 'status' in image and checksum:
        if image['status'] == 'active':
            if 'checksum' not in image:
                # Refresh our info about the image
                image = __salt__['glance.image_show'](image['id'])
            if 'checksum' not in image:
                if not __opts__['test']:
                    ret['result'] = False
                else:
                    ret['result'] = None
                ret['comment'] += 'No checksum available for this image:\n' +\
                        '\tImage has status "{0}".'.format(image['status'])
            elif image['checksum'] != checksum:
                if not __opts__['test']:
                    ret['result'] = False
                else:
                    ret['result'] = None
                ret['comment'] += '"checksum" is {0}, should be {1}.\n'.format(
                    image['checksum'], checksum)
            else:
                ret['comment'] += '"checksum" is correct ({0}).\n'.format(
                    checksum)
        elif image['status'] in ['saving', 'queued']:
            ret['comment'] += 'Checksum won\'t be verified as image ' +\
                'hasn\'t reached\n\t "status=active" yet.\n'
    log.debug('glance.image_present will return: {0}'.format(ret))
    return ret



# -*- coding: utf-8 -*-
'''
ACME / Let's Encrypt module
===========================

.. versionadded: 2016.3

This module currently looks for certbot script in the $PATH as
- certbot,
- lestsencrypt,
- certbot-auto,
- letsencrypt-auto
eventually falls back to /opt/letsencrypt/letsencrypt-auto

.. note::

    Installation & configuration of the Let's Encrypt client can for example be done using
    https://github.com/saltstack-formulas/letsencrypt-formula

.. warning::

    Be sure to set at least accept-tos = True in cli.ini!

Most parameters will fall back to cli.ini defaults if None is given.

'''
# Import python libs
from __future__ import absolute_import, print_function, unicode_literals
import logging
import datetime
import os

# Import salt libs
import salt.utils.path

log = logging.getLogger(__name__)

LEA = salt.utils.path.which_bin(['certbot', 'letsencrypt',
                            'certbot-auto', 'letsencrypt-auto',
                            '/opt/letsencrypt/letsencrypt-auto'])
LE_LIVE = '/etc/letsencrypt/live/'


def __virtual__():
    '''
    Only work when letsencrypt-auto is installed
    '''
    return LEA is not None, 'The ACME execution module cannot be loaded: letsencrypt-auto not installed.'


def _cert_file(name, cert_type):
    '''
    Return expected path of a Let's Encrypt live cert
    '''
    return os.path.join(LE_LIVE, name, '{0}.pem'.format(cert_type))


def _expires(name):
    '''
    Return the expiry date of a cert

    :return datetime object of expiry date
    '''
    cert_file = _cert_file(name, 'cert')
    # Use the salt module if available
    if 'tls.cert_info' in __salt__:
        expiry = __salt__['tls.cert_info'](cert_file)['not_after']
    # Cobble it together using the openssl binary
    else:
        openssl_cmd = 'openssl x509 -in {0} -noout -enddate'.format(cert_file)
        # No %e format on my Linux'es here
        strptime_sux_cmd = 'date --date="$({0} | cut -d= -f2)" +%s'.format(openssl_cmd)
        expiry = float(__salt__['cmd.shell'](strptime_sux_cmd, output_loglevel='quiet'))
        # expiry = datetime.datetime.strptime(expiry.split('=', 1)[-1], '%b %e %H:%M:%S %Y %Z')

    return datetime.datetime.fromtimestamp(expiry)


def _renew_by(name, window=None):
    '''
    Date before a certificate should be renewed

    :param name: Common Name of the certificate (DNS name of certificate)
    :param window: days before expiry date to renew
    :return datetime object of first renewal date
    '''
    expiry = _expires(name)
    if window is not None:
        expiry = expiry - datetime.timedelta(days=window)

    return expiry


def cert(name,
         aliases=None,
         email=None,
         webroot=None,
         test_cert=False,
         renew=None,
         keysize=None,
         server=None,
         owner='root',
         group='root',
         mode='0640',
         certname=None):
    '''
    Obtain/renew a certificate from an ACME CA, probably Let's Encrypt.

    :param name: Common Name of the certificate (DNS name of certificate)
    :param aliases: subjectAltNames (Additional DNS names on certificate)
    :param email: e-mail address for interaction with ACME provider
    :param webroot: True or a full path to use to use webroot. Otherwise use standalone mode
    :param test_cert: Request a certificate from the Happy Hacker Fake CA (mutually exclusive with 'server')
    :param renew: True/'force' to force a renewal, or a window of renewal before expiry in days
    :param keysize: RSA key bits
    :param server: API endpoint to talk to
    :param owner: owner of the private key file
    :param group: group of the private key file
    :param mode: mode of the private key file
    :param certname: Name of the certificate to save
    :return: dict with 'result' True/False/None, 'comment' and certificate's expiry date ('not_after')

    CLI example:

    .. code-block:: bash

        salt 'gitlab.example.com' acme.cert dev.example.com "[gitlab.example.com]" test_cert=True renew=14 webroot=/opt/gitlab/embedded/service/gitlab-rails/public
    '''

    cmd = [LEA, 'certonly', '--non-interactive']

    cert_file = _cert_file(name, 'cert')
    if not __salt__['file.file_exists'](cert_file):
        log.debug('Certificate %s does not exist (yet)', cert_file)
        renew = False
    elif needs_renewal(name, renew):
        log.debug('Certificate %s will be renewed', cert_file)
        cmd.append('--renew-by-default')
        renew = True
    if server:
        cmd.append('--server {0}'.format(server))

    if certname:
        cmd.append('--cert-name {0}'.format(certname))

    if test_cert:
        if server:
            return {'result': False, 'comment': 'Use either server or test_cert, not both'}
        cmd.append('--test-cert')

    if webroot:
        cmd.append('--authenticator webroot')
        if webroot is not True:
            cmd.append('--webroot-path {0}'.format(webroot))
    else:
        cmd.append('--authenticator standalone')

    if email:
        cmd.append('--email {0}'.format(email))

    if keysize:
        cmd.append('--rsa-key-size {0}'.format(keysize))

    cmd.append('--domains {0}'.format(name))
    if aliases is not None:
        for dns in aliases:
            cmd.append('--domains {0}'.format(dns))

    res = __salt__['cmd.run_all'](' '.join(cmd))

    if res['retcode'] != 0:
        return {'result': False, 'comment': 'Certificate {0} renewal failed with:\n{1}'.format(name, res['stderr'])}

    if 'no action taken' in res['stdout']:
        comment = 'Certificate {0} unchanged'.format(cert_file)
    elif renew:
        comment = 'Certificate {0} renewed'.format(name)
    else:
        comment = 'Certificate {0} obtained'.format(name)

    ret = {'comment': comment, 'not_after': expires(name), 'changes': {}, 'result': True}
    ret, _ = __salt__['file.check_perms'](_cert_file(name, 'privkey'),
                                          ret,
                                          owner, group, mode,
                                          follow_symlinks=True)

    return ret


def certs():
    '''
    Return a list of active certificates

    CLI example:

    .. code-block:: bash

        salt 'vhost.example.com' acme.certs
    '''
    return __salt__['file.readdir'](LE_LIVE)[2:]


def info(name):
    '''
    Return information about a certificate

    .. note::
        Will output tls.cert_info if that's available, or OpenSSL text if not

    :param name: CommonName of cert

    CLI example:

    .. code-block:: bash

        salt 'gitlab.example.com' acme.info dev.example.com
    '''
    cert_file = _cert_file(name, 'cert')
    # Use the salt module if available
    if 'tls.cert_info' in __salt__:
        info = __salt__['tls.cert_info'](cert_file)
        # Strip out the extensions object contents;
        # these trip over our poor state output
        # and they serve no real purpose here anyway
        info['extensions'] = info['extensions'].keys()
        return info
    # Cobble it together using the openssl binary
    else:
        openssl_cmd = 'openssl x509 -in {0} -noout -text'.format(cert_file)
        return __salt__['cmd.run'](openssl_cmd, output_loglevel='quiet')


def expires(name):
    '''
    The expiry date of a certificate in ISO format

    :param name: CommonName of cert

    CLI example:

    .. code-block:: bash

        salt 'gitlab.example.com' acme.expires dev.example.com
    '''
    return _expires(name).isoformat()


def has(name):
    '''
    Test if a certificate is in the Let's Encrypt Live directory

    :param name: CommonName of cert

    Code example:

    .. code-block:: python

        if __salt__['acme.has']('dev.example.com'):
            log.info('That is one nice certificate you have there!')
    '''
    return __salt__['file.file_exists'](_cert_file(name, 'cert'))


def renew_by(name, window=None):
    '''
    Date in ISO format when a certificate should first be renewed

    :param name: CommonName of cert
    :param window: number of days before expiry when renewal should take place
    '''
    return _renew_by(name, window).isoformat()


def needs_renewal(name, window=None):
    '''
    Check if a certificate needs renewal

    :param name: CommonName of cert
    :param window: Window in days to renew earlier or True/force to just return True

    Code example:

    .. code-block:: python

        if __salt__['acme.needs_renewal']('dev.example.com'):
            __salt__['acme.cert']('dev.example.com', **kwargs)
        else:
            log.info('Your certificate is still good')
    '''
    if window is not None and window in ('force', 'Force', True):
        return True

    return _renew_by(name, window) <= datetime.datetime.today()



# -*- coding: utf-8 -*-
'''
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    salt.utils.vt
    ~~~~~~~~~~~~~

    Virtual Terminal

    This code has been heavily inspired by Python's subprocess code, the `non
    blocking version of it`__, some minor online snippets about TTY handling
    with python including `Python's own ``pty`` source code`__ and `Pexpect`__
    which has already surpassed some of the pitfalls that some systems would
    get us into.

    .. __: http://code.activestate.com/recipes/440554/
    .. __: https://github.com/python-mirror/python/blob/3.3/Lib/pty.py
    .. __: https://github.com/pexpect/pexpect

'''
from __future__ import absolute_import, print_function, unicode_literals

# Import python libs
import os
import sys
import time
import errno
import signal
import select
import logging

# Import salt libs
from salt.ext import six

mswindows = (sys.platform == "win32")

try:
    # pylint: disable=F0401,W0611
    from win32file import ReadFile, WriteFile
    from win32pipe import PeekNamedPipe
    import msvcrt
    import win32api
    import win32con
    import win32process
    # pylint: enable=F0401,W0611
except ImportError:
    import pty
    import fcntl
    import struct
    import termios
    import resource

# Import salt libs
import salt.utils.crypt
import salt.utils.stringutils
from salt.ext.six import string_types
from salt.log.setup import LOG_LEVELS

log = logging.getLogger(__name__)


class TerminalException(Exception):
    '''
    Terminal specific exception
    '''


# ----- Cleanup Running Instances ------------------------------------------->
# This lists holds Terminal instances for which the underlying process had
# not exited at the time its __del__ method got called: those processes are
# wait()ed for synchronously from _cleanup() when a new Terminal object is
# created, to avoid zombie processes.
_ACTIVE = []


def _cleanup():
    '''
    Make sure that any terminal processes still running when __del__ was called
    to the waited and cleaned up.
    '''
    for inst in _ACTIVE[:]:
        res = inst.isalive()
        if res is not True:
            try:
                _ACTIVE.remove(inst)
            except ValueError:
                # This can happen if two threads create a new Terminal instance
                # It's harmless that it was already removed, so ignore.
                pass
# <---- Cleanup Running Instances --------------------------------------------


class Terminal(object):
    '''
    I'm a virtual terminal
    '''
    def __init__(self,
                 args=None,
                 executable=None,
                 shell=False,
                 cwd=None,
                 env=None,
                 preexec_fn=None,

                 # Terminal Size
                 rows=None,
                 cols=None,

                 # Logging options
                 log_stdin=None,
                 log_stdin_level='debug',
                 log_stdout=None,
                 log_stdout_level='debug',
                 log_stderr=None,
                 log_stderr_level='debug',

                 # sys.stdXYZ streaming options
                 stream_stdout=None,
                 stream_stderr=None,
                 ):

        # Let's avoid Zombies!!!
        _cleanup()

        if not args and not executable:
            raise TerminalException(
                'You need to pass at least one of "args", "executable" '
            )

        self.args = args
        self.executable = executable
        self.shell = shell
        self.cwd = cwd
        self.env = env
        self.preexec_fn = preexec_fn

        # ----- Set the desired terminal size ------------------------------->
        if rows is None and cols is None:
            rows, cols = self.__detect_parent_terminal_size()
        elif rows is not None and cols is None:
            _, cols = self.__detect_parent_terminal_size()
        elif rows is None and cols is not None:
            rows, _ = self.__detect_parent_terminal_size()
        self.rows = rows
        self.cols = cols
        # <---- Set the desired terminal size --------------------------------

        # ----- Internally Set Attributes ----------------------------------->
        self.pid = None
        self.stdin = None
        self.stdout = None
        self.stderr = None

        self.child_fd = None
        self.child_fde = None

        self.closed = True
        self.flag_eof_stdout = False
        self.flag_eof_stderr = False
        self.terminated = True
        self.exitstatus = None
        self.signalstatus = None
        # status returned by os.waitpid
        self.status = None
        self.__irix_hack = 'irix' in sys.platform.lower()
        # <---- Internally Set Attributes ------------------------------------

        # ----- Direct Streaming Setup -------------------------------------->
        if stream_stdout is True:
            self.stream_stdout = sys.stdout
        elif stream_stdout is False:
            self.stream_stdout = None
        elif stream_stdout is not None:
            if not hasattr(stream_stdout, 'write') or \
                    not hasattr(stream_stdout, 'flush') or \
                    not hasattr(stream_stdout, 'close'):
                raise TerminalException(
                    '\'stream_stdout\' needs to have at least 3 methods, '
                    '\'write()\', \'flush()\' and \'close()\'.'
                )
            self.stream_stdout = stream_stdout
        else:
            raise TerminalException(
                'Don\'t know how to handle \'{0}\' as the VT\'s '
                '\'stream_stdout\' parameter.'.format(stream_stdout)
            )

        if stream_stderr is True:
            self.stream_stderr = sys.stderr
        elif stream_stderr is False:
            self.stream_stderr = None
        elif stream_stderr is not None:
            if not hasattr(stream_stderr, 'write') or \
                    not hasattr(stream_stderr, 'flush') or \
                    not hasattr(stream_stderr, 'close'):
                raise TerminalException(
                    '\'stream_stderr\' needs to have at least 3 methods, '
                    '\'write()\', \'flush()\' and \'close()\'.'
                )
            self.stream_stderr = stream_stderr
        else:
            raise TerminalException(
                'Don\'t know how to handle \'{0}\' as the VT\'s '
                '\'stream_stderr\' parameter.'.format(stream_stderr)
            )
        # <---- Direct Streaming Setup ---------------------------------------

        # ----- Spawn our terminal ------------------------------------------>
        try:
            self._spawn()
        except Exception as err:  # pylint: disable=W0703
            # A lot can go wrong, so that's why we're catching the most general
            # exception type
            log.warning(
                'Failed to spawn the VT: %s', err,
                 exc_info_on_loglevel=logging.DEBUG
            )
            raise TerminalException(
                'Failed to spawn the VT. Error: {0}'.format(err)
            )

        log.debug(
            'Child Forked! PID: %s  STDOUT_FD: %s  STDERR_FD: %s',
            self.pid, self.child_fd, self.child_fde
        )
        terminal_command = ' '.join(self.args)
        if 'decode("base64")' in terminal_command or 'base64.b64decode(' in terminal_command:
            log.debug('VT: Salt-SSH SHIM Terminal Command executed. Logged to TRACE')
            log.trace('Terminal Command: %s', terminal_command)
        else:
            log.debug('Terminal Command: %s', terminal_command)
        # <---- Spawn our terminal -------------------------------------------

        # ----- Setup Logging ----------------------------------------------->
        # Setup logging after spawned in order to have a pid value
        self.stdin_logger_level = LOG_LEVELS.get(log_stdin_level, log_stdin_level)
        if log_stdin is True:
            self.stdin_logger = logging.getLogger(
                '{0}.{1}.PID-{2}.STDIN'.format(
                    __name__, self.__class__.__name__, self.pid
                )
            )
        elif log_stdin is not None:
            if not isinstance(log_stdin, logging.Logger):
                raise RuntimeError(
                    '\'log_stdin\' needs to subclass `logging.Logger`'
                )
            self.stdin_logger = log_stdin
        else:
            self.stdin_logger = None

        self.stdout_logger_level = LOG_LEVELS.get(log_stdout_level, log_stdout_level)
        if log_stdout is True:
            self.stdout_logger = logging.getLogger(
                '{0}.{1}.PID-{2}.STDOUT'.format(
                    __name__, self.__class__.__name__, self.pid
                )
            )
        elif log_stdout is not None:
            if not isinstance(log_stdout, logging.Logger):
                raise RuntimeError(
                    '\'log_stdout\' needs to subclass `logging.Logger`'
                )
            self.stdout_logger = log_stdout
        else:
            self.stdout_logger = None

        self.stderr_logger_level = LOG_LEVELS.get(log_stderr_level, log_stderr_level)
        if log_stderr is True:
            self.stderr_logger = logging.getLogger(
                '{0}.{1}.PID-{2}.STDERR'.format(
                    __name__, self.__class__.__name__, self.pid
                )
            )
        elif log_stderr is not None:
            if not isinstance(log_stderr, logging.Logger):
                raise RuntimeError(
                    '\'log_stderr\' needs to subclass `logging.Logger`'
                )
            self.stderr_logger = log_stderr
        else:
            self.stderr_logger = None
        # <---- Setup Logging ------------------------------------------------

    # ----- Common Public API ----------------------------------------------->
    def send(self, data):
        '''
        Send data to the terminal. You are responsible to send any required
        line feeds.
        '''
        return self._send(data)

    def sendline(self, data, linesep=os.linesep):
        '''
        Send the provided data to the terminal appending a line feed.
        '''
        return self.send('{0}{1}'.format(data, linesep))

    def recv(self, maxsize=None):
        '''
        Receive data from the terminal as a (``stdout``, ``stderr``) tuple. If
        any of those is ``None`` we can no longer communicate with the
        terminal's child process.
        '''
        if maxsize is None:
            maxsize = 1024
        elif maxsize < 1:
            maxsize = 1
        return self._recv(maxsize)

    def close(self, terminate=True, kill=False):
        '''
        Close the communication with the terminal's child.
        If ``terminate`` is ``True`` then additionally try to terminate the
        terminal, and if ``kill`` is also ``True``, kill the terminal if
        terminating it was not enough.
        '''
        if not self.closed:
            if self.child_fd is not None:
                os.close(self.child_fd)
                self.child_fd = None
            if self.child_fde is not None:
                os.close(self.child_fde)
                self.child_fde = None
            time.sleep(0.1)
            if terminate:
                if not self.terminate(kill):
                    raise TerminalException('Failed to terminate child process.')
            self.closed = True

    @property
    def has_unread_data(self):
        return self.flag_eof_stderr is False or self.flag_eof_stdout is False

    # <---- Common Public API ------------------------------------------------

    # ----- Common Internal API --------------------------------------------->
    def _translate_newlines(self, data):
        if data is None or not data:
            return
        # PTY's always return \r\n as the line feeds
        return data.replace('\r\n', os.linesep)
    # <---- Common Internal API ----------------------------------------------

    # ----- Context Manager Methods ----------------------------------------->
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close(terminate=True, kill=True)
        # Wait for the process to terminate, to avoid zombies.
        if self.isalive():
            self.wait()
    # <---- Context Manager Methods ------------------------------------------

# ----- Platform Specific Methods ------------------------------------------->
    if mswindows:
        # ----- Windows Methods --------------------------------------------->
        def _execute(self):
            raise NotImplementedError

        def _spawn(self):
            raise NotImplementedError

        def _recv(self, maxsize):
            raise NotImplementedError

        def _send(self, data):
            raise NotImplementedError

        def send_signal(self, sig):
            '''
            Send a signal to the process
            '''
            # pylint: disable=E1101
            if sig == signal.SIGTERM:
                self.terminate()
            elif sig == signal.CTRL_C_EVENT:
                os.kill(self.pid, signal.CTRL_C_EVENT)
            elif sig == signal.CTRL_BREAK_EVENT:
                os.kill(self.pid, signal.CTRL_BREAK_EVENT)
            else:
                raise ValueError('Unsupported signal: {0}'.format(sig))
            # pylint: enable=E1101

        def terminate(self):
            '''
            Terminates the process
            '''
            try:
                win32api.TerminateProcess(self._handle, 1)
            except OSError:
                # ERROR_ACCESS_DENIED (winerror 5) is received when the
                # process already died.
                ecode = win32process.GetExitCodeProcess(self._handle)
                if ecode == win32con.STILL_ACTIVE:
                    raise
                self.exitstatus = ecode

        kill = terminate
    # <---- Windows Methods --------------------------------------------------
    else:
        # ----- Linux Methods ----------------------------------------------->
        # ----- Internal API ------------------------------------------------>
        def _spawn(self):
            self.pid, self.child_fd, self.child_fde = self.__fork_ptys()

            if isinstance(self.args, string_types):
                args = [self.args]
            elif self.args:
                args = list(self.args)
            else:
                args = []

            if self.shell and self.args:
                self.args = ['/bin/sh', '-c', ' '.join(args)]
            elif self.shell:
                self.args = ['/bin/sh']
            else:
                self.args = args

            if self.executable:
                self.args[0] = self.executable

            if self.executable is None:
                self.executable = self.args[0]

            if self.pid == 0:
                # Child
                self.stdin = sys.stdin.fileno()
                self.stdout = sys.stdout.fileno()
                self.stderr = sys.stderr.fileno()

                # Set the terminal size
                self.child_fd = self.stdin

                if os.isatty(self.child_fd):
                    # Only try to set the window size if the parent IS a tty
                    try:
                        self.setwinsize(self.rows, self.cols)
                    except IOError as err:
                        log.warning(
                            'Failed to set the VT terminal size: %s',
                            err, exc_info_on_loglevel=logging.DEBUG
                        )

                # Do not allow child to inherit open file descriptors from
                # parent
                max_fd = resource.getrlimit(resource.RLIMIT_NOFILE)
                try:
                    os.closerange(pty.STDERR_FILENO + 1, max_fd[0])
                except OSError:
                    pass

                if self.cwd is not None:
                    os.chdir(self.cwd)

                if self.preexec_fn:
                    self.preexec_fn()

                if self.env is None:
                    os.execvp(self.executable, self.args)
                else:
                    os.execvpe(self.executable, self.args, self.env)

            # Parent
            self.closed = False
            self.terminated = False

        def __fork_ptys(self):
            '''
            Fork the PTY

            The major difference from the python source is that we separate the
            stdout from stderr output.
            '''
            stdout_parent_fd, stdout_child_fd = pty.openpty()
            if stdout_parent_fd < 0 or stdout_child_fd < 0:
                raise TerminalException('Failed to open a TTY for stdout')

            stderr_parent_fd, stderr_child_fd = pty.openpty()
            if stderr_parent_fd < 0 or stderr_child_fd < 0:
                raise TerminalException('Failed to open a TTY for stderr')

            pid = os.fork()
            if pid < pty.CHILD:
                raise TerminalException('Failed to fork')
            elif pid == pty.CHILD:
                # Child.
                # Close parent FDs
                os.close(stdout_parent_fd)
                os.close(stderr_parent_fd)
                salt.utils.crypt.reinit_crypto()

                # ----- Make STDOUT the controlling PTY --------------------->
                child_name = os.ttyname(stdout_child_fd)
                # Disconnect from controlling tty. Harmless if not already
                # connected
                try:
                    tty_fd = os.open('/dev/tty', os.O_RDWR | os.O_NOCTTY)
                    if tty_fd >= 0:
                        os.close(tty_fd)
                # which exception, shouldn't we catch explicitly .. ?
                except:  # pylint: disable=W0702
                    # Already disconnected. This happens if running inside cron
                    pass

                # New session!
                os.setsid()

                # Verify we are disconnected from controlling tty
                # by attempting to open it again.
                try:
                    tty_fd = os.open('/dev/tty', os.O_RDWR | os.O_NOCTTY)
                    if tty_fd >= 0:
                        os.close(tty_fd)
                        raise TerminalException(
                            'Failed to disconnect from controlling tty. It is '
                            'still possible to open /dev/tty.'
                        )
                # which exception, shouldn't we catch explicitly .. ?
                except:  # pylint: disable=W0702
                    # Good! We are disconnected from a controlling tty.
                    pass

                # Verify we can open child pty.
                tty_fd = os.open(child_name, os.O_RDWR)
                if tty_fd < 0:
                    raise TerminalException(
                        'Could not open child pty, {0}'.format(child_name)
                    )
                else:
                    os.close(tty_fd)

                # Verify we now have a controlling tty.
                if os.name != 'posix':
                    # Only do this check in not BSD-like operating systems. BSD-like operating systems breaks at this point
                    tty_fd = os.open('/dev/tty', os.O_WRONLY)
                    if tty_fd < 0:
                        raise TerminalException(
                            'Could not open controlling tty, /dev/tty'
                        )
                    else:
                        os.close(tty_fd)
                # <---- Make STDOUT the controlling PTY ----------------------

                # ----- Duplicate Descriptors ------------------------------->
                os.dup2(stdout_child_fd, pty.STDIN_FILENO)
                os.dup2(stdout_child_fd, pty.STDOUT_FILENO)
                os.dup2(stderr_child_fd, pty.STDERR_FILENO)
                # <---- Duplicate Descriptors --------------------------------
            else:
                # Parent. Close Child PTY's
                salt.utils.crypt.reinit_crypto()
                os.close(stdout_child_fd)
                os.close(stderr_child_fd)

            return pid, stdout_parent_fd, stderr_parent_fd

        def _send(self, data):
            if self.child_fd is None:
                return None

            if not select.select([], [self.child_fd], [], 0)[1]:
                return 0

            try:
                if self.stdin_logger:
                    self.stdin_logger.log(self.stdin_logger_level, data)
                if six.PY3:
                    written = os.write(self.child_fd, data.encode(__salt_system_encoding__))
                else:
                    written = os.write(self.child_fd, data)
            except OSError as why:
                if why.errno == errno.EPIPE:  # broken pipe
                    os.close(self.child_fd)
                    self.child_fd = None
                    return
                raise
            return written

        def _recv(self, maxsize):
            rfds = []
            if self.child_fd:
                rfds.append(self.child_fd)
            if self.child_fde:
                rfds.append(self.child_fde)

            if not self.isalive():
                if not rfds:
                    return None, None
                rlist, _, _ = select.select(rfds, [], [], 0)
                if not rlist:
                    self.flag_eof_stdout = self.flag_eof_stderr = True
                    log.debug('End of file(EOL). Brain-dead platform.')
                    return None, None
            elif self.__irix_hack:
                # Irix takes a long time before it realizes a child was
                # terminated.
                # FIXME So does this mean Irix systems are forced to always
                # have a 2 second delay when calling read_nonblocking?
                # That sucks.
                rlist, _, _ = select.select(rfds, [], [], 2)
                if not rlist:
                    self.flag_eof_stdout = self.flag_eof_stderr = True
                    log.debug('End of file(EOL). Slow platform.')
                    return None, None

            stderr = ''
            stdout = ''

            # ----- Store FD Flags ------------------------------------------>
            if self.child_fd:
                fd_flags = fcntl.fcntl(self.child_fd, fcntl.F_GETFL)
            if self.child_fde:
                fde_flags = fcntl.fcntl(self.child_fde, fcntl.F_GETFL)
            # <---- Store FD Flags -------------------------------------------

            # ----- Non blocking Reads -------------------------------------->
            if self.child_fd:
                fcntl.fcntl(self.child_fd,
                            fcntl.F_SETFL, fd_flags | os.O_NONBLOCK)
            if self.child_fde:
                fcntl.fcntl(self.child_fde,
                            fcntl.F_SETFL, fde_flags | os.O_NONBLOCK)
            # <---- Non blocking Reads ---------------------------------------

            # ----- Check for any incoming data ----------------------------->
            rlist, _, _ = select.select(rfds, [], [], 0)
            # <---- Check for any incoming data ------------------------------

            # ----- Nothing to Process!? ------------------------------------>
            if not rlist:
                if not self.isalive():
                    self.flag_eof_stdout = self.flag_eof_stderr = True
                    log.debug('End of file(EOL). Very slow platform.')
                    return None, None
            # <---- Nothing to Process!? -------------------------------------

            # ----- Process STDERR ------------------------------------------>
            if self.child_fde in rlist:
                try:
                    stderr = self._translate_newlines(
                        salt.utils.stringutils.to_unicode(
                            os.read(self.child_fde, maxsize)
                        )
                    )

                    if not stderr:
                        self.flag_eof_stderr = True
                        stderr = None
                    else:
                        if self.stream_stderr:
                            self.stream_stderr.write(stderr)
                            self.stream_stderr.flush()

                        if self.stderr_logger:
                            stripped = stderr.rstrip()
                            if stripped.startswith(os.linesep):
                                stripped = stripped[len(os.linesep):]
                            if stripped:
                                self.stderr_logger.log(self.stderr_logger_level, stripped)
                except OSError:
                    os.close(self.child_fde)
                    self.child_fde = None
                    self.flag_eof_stderr = True
                    stderr = None
                finally:
                    if self.child_fde is not None:
                        fcntl.fcntl(self.child_fde, fcntl.F_SETFL, fde_flags)
            # <---- Process STDERR -------------------------------------------

            # ----- Process STDOUT ------------------------------------------>
            if self.child_fd in rlist:
                try:
                    stdout = self._translate_newlines(
                        salt.utils.stringutils.to_unicode(
                            os.read(self.child_fd, maxsize)
                        )
                    )

                    if not stdout:
                        self.flag_eof_stdout = True
                        stdout = None
                    else:
                        if self.stream_stdout:
                            self.stream_stdout.write(stdout)
                            self.stream_stdout.flush()

                        if self.stdout_logger:
                            stripped = stdout.rstrip()
                            if stripped.startswith(os.linesep):
                                stripped = stripped[len(os.linesep):]
                            if stripped:
                                self.stdout_logger.log(self.stdout_logger_level, stripped)
                except OSError:
                    os.close(self.child_fd)
                    self.child_fd = None
                    self.flag_eof_stdout = True
                    stdout = None
                finally:
                    if self.child_fd is not None:
                        fcntl.fcntl(self.child_fd, fcntl.F_SETFL, fd_flags)
            # <---- Process STDOUT -------------------------------------------
            return stdout, stderr

        def __detect_parent_terminal_size(self):
            try:
                TIOCGWINSZ = getattr(termios, 'TIOCGWINSZ', 1074295912)
                packed = struct.pack(b'HHHH', 0, 0, 0, 0)
                ioctl = fcntl.ioctl(sys.stdin.fileno(), TIOCGWINSZ, packed)
                return struct.unpack(b'HHHH', ioctl)[0:2]
            except IOError:
                # Return a default value of 24x80
                return 24, 80
        # <---- Internal API -------------------------------------------------

        # ----- Public API -------------------------------------------------->
        def getwinsize(self):
            '''
            This returns the terminal window size of the child tty. The return
            value is a tuple of (rows, cols).

            Thank you for the shortcut PEXPECT
            '''
            if self.child_fd is None:
                raise TerminalException(
                    'Can\'t check the size of the terminal since we\'re not '
                    'connected to the child process.'
                )

            TIOCGWINSZ = getattr(termios, 'TIOCGWINSZ', 1074295912)
            packed = struct.pack(b'HHHH', 0, 0, 0, 0)
            ioctl = fcntl.ioctl(self.child_fd, TIOCGWINSZ, packed)
            return struct.unpack(b'HHHH', ioctl)[0:2]

        def setwinsize(self, rows, cols):
            '''
            This sets the terminal window size of the child tty. This will
            cause a SIGWINCH signal to be sent to the child. This does not
            change the physical window size. It changes the size reported to
            TTY-aware applications like vi or curses -- applications that
            respond to the SIGWINCH signal.

            Thank you for the shortcut PEXPECT
            '''
            # Check for buggy platforms. Some Python versions on some platforms
            # (notably OSF1 Alpha and RedHat 7.1) truncate the value for
            # termios.TIOCSWINSZ. It is not clear why this happens.
            # These platforms don't seem to handle the signed int very well;
            # yet other platforms like OpenBSD have a large negative value for
            # TIOCSWINSZ and they don't have a truncate problem.
            # Newer versions of Linux have totally different values for
            # TIOCSWINSZ.
            # Note that this fix is a hack.
            TIOCSWINSZ = getattr(termios, 'TIOCSWINSZ', -2146929561)
            if TIOCSWINSZ == 2148037735:
                # Same bits, but with sign.
                TIOCSWINSZ = -2146929561
            # Note, assume ws_xpixel and ws_ypixel are zero.
            packed = struct.pack(b'HHHH', rows, cols, 0, 0)
            fcntl.ioctl(self.child_fd, TIOCSWINSZ, packed)

        def isalive(self,
                    _waitpid=os.waitpid,
                    _wnohang=os.WNOHANG,
                    _wifexited=os.WIFEXITED,
                    _wexitstatus=os.WEXITSTATUS,
                    _wifsignaled=os.WIFSIGNALED,
                    _wifstopped=os.WIFSTOPPED,
                    _wtermsig=os.WTERMSIG,
                    _os_error=os.error,
                    _errno_echild=errno.ECHILD,
                    _terminal_exception=TerminalException):
            '''
            This tests if the child process is running or not. This is
            non-blocking. If the child was terminated then this will read the
            exitstatus or signalstatus of the child. This returns True if the
            child process appears to be running or False if not. It can take
            literally SECONDS for Solaris to return the right status.
            '''
            if self.terminated:
                return False

            if self.has_unread_data is False:
                # This is for Linux, which requires the blocking form
                # of waitpid to get status of a defunct process.
                # This is super-lame. The flag_eof_* would have been set
                # in recv(), so this should be safe.
                waitpid_options = 0
            else:
                waitpid_options = _wnohang

            try:
                pid, status = _waitpid(self.pid, waitpid_options)
            except _os_error:
                err = sys.exc_info()[1]
                # No child processes
                if err.errno == _errno_echild:
                    raise _terminal_exception(
                        'isalive() encountered condition where "terminated" '
                        'is 0, but there was no child process. Did someone '
                        'else call waitpid() on our process?'
                    )
                else:
                    raise err

            # I have to do this twice for Solaris.
            # I can't even believe that I figured this out...
            # If waitpid() returns 0 it means that no child process
            # wishes to report, and the value of status is undefined.
            if pid == 0:
                try:
                    ### os.WNOHANG # Solaris!
                    pid, status = _waitpid(self.pid, waitpid_options)
                except _os_error as exc:
                    # This should never happen...
                    if exc.errno == _errno_echild:
                        raise _terminal_exception(
                            'isalive() encountered condition that should '
                            'never happen. There was no child process. Did '
                            'someone else call waitpid() on our process?'
                        )
                    else:
                        raise

                # If pid is still 0 after two calls to waitpid() then the
                # process really is alive. This seems to work on all platforms,
                # except for Irix which seems to require a blocking call on
                # waitpid or select, so I let recv take care of this situation
                # (unfortunately, this requires waiting through the timeout).
                if pid == 0:
                    return True

            if pid == 0:
                return True

            if _wifexited(status):
                self.status = status
                self.exitstatus = _wexitstatus(status)
                self.signalstatus = None
                self.terminated = True
            elif _wifsignaled(status):
                self.status = status
                self.exitstatus = None
                self.signalstatus = _wtermsig(status)
                self.terminated = True
            elif _wifstopped(status):
                raise _terminal_exception(
                    'isalive() encountered condition where child process is '
                    'stopped. This is not supported. Is some other process '
                    'attempting job control with our child pid?'
                )
            return False

        def terminate(self, force=False):
            '''
            This forces a child process to terminate. It starts nicely with
            SIGHUP and SIGINT. If "force" is True then moves onto SIGKILL. This
            returns True if the child was terminated. This returns False if the
            child could not be terminated.
            '''
            if not self.closed:
                self.close(terminate=False)

            if not self.isalive():
                return True
            try:
                self.send_signal(signal.SIGHUP)
                time.sleep(0.1)
                if not self.isalive():
                    return True
                self.send_signal(signal.SIGCONT)
                time.sleep(0.1)
                if not self.isalive():
                    return True
                self.send_signal(signal.SIGINT)
                time.sleep(0.1)
                if not self.isalive():
                    return True
                if force:
                    self.send_signal(signal.SIGKILL)
                    time.sleep(0.1)
                    if not self.isalive():
                        return True
                    else:
                        return False
                return False
            except OSError:
                # I think there are kernel timing issues that sometimes cause
                # this to happen. I think isalive() reports True, but the
                # process is dead to the kernel.
                # Make one last attempt to see if the kernel is up to date.
                time.sleep(0.1)
                if not self.isalive():
                    return True
                else:
                    return False

        def wait(self):
            '''
            This waits until the child exits internally consuming any remaining
            output from the child, thus, no blocking forever because the child
            has unread data.
            '''
            if self.isalive():
                while self.isalive():
                    stdout, stderr = self.recv()
                    if stdout is None:
                        break
                    if stderr is None:
                        break
            else:
                raise TerminalException('Cannot wait for dead child process.')
            return self.exitstatus

        def send_signal(self, sig):
            '''
            Send a signal to the process
            '''
            os.kill(self.pid, sig)

        def kill(self):
            '''
            Kill the process with SIGKILL
            '''
            self.send_signal(signal.SIGKILL)
        # <---- Public API ---------------------------------------------------
    # <---- Linux Methods ----------------------------------------------------

    # ----- Cleanup!!! ------------------------------------------------------>
    def __del__(self, _maxsize=sys.maxsize, _active=_ACTIVE):  # pylint: disable=W0102
        # I've disabled W0102 above which is regarding a dangerous default
        # value of [] for _ACTIVE, though, this is how Python itself handles
        # their subprocess clean up code.
        # XXX: Revisit this cleanup code to make it less dangerous.

        if self.pid is None:
            # We didn't get to successfully create a child process.
            return

        # In case the child hasn't been waited on, check if it's done.
        if self.isalive() and _ACTIVE is not None:
            # Child is still running, keep us alive until we can wait on it.
            _ACTIVE.append(self)
    # <---- Cleanup!!! -------------------------------------------------------
# <---- Platform Specific Methods --------------------------------------------



# -*- coding: utf-8 -*-

# pylint: disable=W1401
'''
This module renders highstate configuration into a more human readable format.

How it works:

 `highstate or lowstate` data is parsed with a `proccesser` this defaults to `highstate_doc.proccesser_markdown`.
 The proccessed data is passed to a `jinja` template that builds up the document content.


configuration: Pillar

.. code-block:: yaml

    # the following defaults can be overrided
    highstate_doc.config:

        # list of regex of state names to ignore in `highstate_doc.proccess_lowstates`
        filter_id_regex:
            - '.*!doc_skip$'

        # list of regex of state functions to ignore in `highstate_doc.proccess_lowstates`
        filter_state_function_regex:
            - 'file.accumulated'

        # dict of regex to replace text after `highstate_doc.render`. (remove passwords)
        text_replace_regex:
            'password:.*^': '[PASSWORD]'

        # limit size of files that can be included in doc (10000 bytes)
        max_render_file_size: 10000

        # advanced option to set a custom lowstate proccesser
        proccesser: highstate_doc.proccesser_markdown


State example

.. code-block:: yaml

    {{sls}} note:
        highstate_doc.note:
            - name: example
            - order: 0
            - contents: |
                example `highstate_doc.note`
                ------------------
                This state does not do anything to the system! It is only used by a `proccesser`
                you can use `requisites` and `order` to move your docs around the rendered file.

    {{sls}} a file we dont want in the doc !doc_skip:
        file.managed:
            - name: /root/passwords
            - contents: 'password: sadefgq34y45h56q'
            # also could use `highstate_doc.config: text_replace_regex` to replace
            # password string. `password:.*^': '[PASSWORD]`


To create the help document build a State that uses `highstate_doc.render`.
For preformance it's advised to not included this state in your `top.sls` file.

.. code-block:: yaml

    # example `salt://makereadme.sls`
    make helpfile:
        file.managed:
            - name: /root/README.md
            - contents: {{salt.highstate_doc.render()|json}}
            - show_diff: {{opts['test']}}
            - mode: '0640'
            - order: last

Run our `makereadme.sls` state to create `/root/README.md`.

.. code-block:: bash

    # first ensure `highstate` return without errors or changes
    salt-call state.highstate
    salt-call state.apply makereadme
    # or if you dont want the extra `make helpfile` state
    salt-call --out=newline_values_only salt.highstate_doc.render > /root/README.md ; chmod 0600 /root/README.md


Creating a document collection
------------------------------

From the master we can run the following script to
creates a collection of all your minion documents.

.. code-block:: bash

    salt '*' state.apply makereadme

.. code-block:: python

    #!/bin/python
    import os
    import salt.client
    s = salt.client.LocalClient()
    # NOTE: because of issues with `cp.push` use `highstate_doc.read_file`
    o = s.cmd('*', 'highstate_doc.read_file', ['/root/README.md'])
    for m in o:
        d = o.get(m)
        if d and not d.endswith('is not available.'):
            # mkdir m
            #directory = os.path.dirname(file_path)
            if not os.path.exists(m):
                os.makedirs(m)
            with open(m + '/README.md','wb') as f:
                f.write(d)
            print('ADDED: ' + m + '/README.md')


Once the master has a collection of all the README files.
You can use pandoc to create HTML versions of the markdown.

.. code-block:: bash

    # proccess all the readme.md files to readme.html
    if which pandoc; then echo "Found pandoc"; else echo "** Missing pandoc"; exit 1; fi
    if which gs; then echo "Found gs"; else echo "** Missing gs(ghostscript)"; exit 1; fi
    readme_files=$(find $dest -type f -path "*/README.md" -print)
    for f in $readme_files ; do
        ff=${f#$dest/}
        minion=${ff%%/*}
        echo "proccess: $dest/${minion}/$(basename $f)"
        cat $dest/${minion}/$(basename $f) | \
            pandoc --standalone --from markdown_github --to html \
            --include-in-header $dest/style.html \
            > $dest/${minion}/$(basename $f).html
    done

It is also nice to put the help files in source control.

    # git init
    git add -A
    git commit -am 'updated docs'
    git push -f


Other hints
-----------

If you wish to customize the document format:

.. code-block:: yaml

    # you could also create a new `proccesser` for perhaps reStructuredText
    # highstate_doc.config:
    #     proccesser: doc_custom.proccesser_rst

    # example `salt://makereadme.jinja`
    """
    {{opts['id']}}
    ==========================================

    {# lowstates is set from highstate_doc.render() #}
    {# if lowstates is missing use salt.highstate_doc.proccess_lowstates() #}
    {% for s in lowstates %}
    {{s.id}}
    -----------------------------------------------------------------
    {{s.function}}

    {{s.markdown.requisite}}
    {{s.markdown.details}}

    {%- endfor %}
    """

    # example `salt://makereadme.sls`
    {% import_text "makereadme.jinja" as makereadme %}
    {{sls}} or:
        file.managed:
            - name: /root/README_other.md
            - contents: {{salt.highstate_doc.render(jinja_template_text=makereadme)|json}}
            - mode: '0640'


Some `replace_text_regex` values that might be helpful.

    ## CERTS
    '-----BEGIN RSA PRIVATE KEY-----[\r\n\t\f\S]{0,2200}': 'XXXXXXX'
    '-----BEGIN CERTIFICATE-----[\r\n\t\f\S]{0,2200}': 'XXXXXXX'
    '-----BEGIN DH PARAMETERS-----[\r\n\t\f\S]{0,2200}': 'XXXXXXX'
    '-----BEGIN PRIVATE KEY-----[\r\n\t\f\S]{0,2200}': 'XXXXXXX'
    '-----BEGIN OPENSSH PRIVATE KEY-----[\r\n\t\f\S]{0,2200}': 'XXXXXXX'
    'ssh-rsa .* ': 'ssh-rsa XXXXXXX '
    'ssh-dss .* ': 'ssh-dss XXXXXXX '
    ## DB
    'DB_PASS.*': 'DB_PASS = XXXXXXX'
    '5432:*:*:.*': '5432:*:XXXXXXX'
    "'PASSWORD': .*": "'PASSWORD': 'XXXXXXX',"
    " PASSWORD '.*'": " PASSWORD 'XXXXXXX'"
    'PGPASSWORD=.* ': 'PGPASSWORD=XXXXXXX'
    "_replication password '.*'":  "_replication password 'XXXXXXX'"
    ## OTHER
    'EMAIL_HOST_PASSWORD =.*': 'EMAIL_HOST_PASSWORD =XXXXXXX'
    "net ads join -U '.*@MFCFADS.MATH.EXAMPLE.CA.* ": "net ads join -U '.*@MFCFADS.MATH.EXAMPLE.CA%XXXXXXX "
    "net ads join -U '.*@NEXUS.EXAMPLE.CA.* ": "net ads join -U '.*@NEXUS.EXAMPLE.CA%XXXXXXX "
    'install-uptrack .* --autoinstall': 'install-uptrack XXXXXXX --autoinstall'
    'accesskey = .*': 'accesskey = XXXXXXX'
    'auth_pass .*': 'auth_pass XXXXXXX'
    'PSK "0x.*': 'PSK "0xXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX'
    'SECRET_KEY.*': 'SECRET_KEY = XXXXXXX'
    "password=.*": "password=XXXXXXX"
    '<password>.*</password>': '<password>XXXXXXX</password>'
    '<salt>.*</salt>': '<salt>XXXXXXX</salt>'
    'application.secret = ".*"': 'application.secret = "XXXXXXX"'
    'url = "postgres://.*"': 'url = "postgres://XXXXXXX"'
    'PASS_.*_PASS': 'PASS_XXXXXXX_PASS'
    ## HTACCESS
    ':{PLAIN}.*': ':{PLAIN}XXXXXXX'

'''

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals
import re
import logging

# Import Salt libs
import salt.utils.files
import salt.utils.stringutils
import salt.utils.templates as tpl
import salt.utils.yaml

__virtualname__ = 'highstate_doc'

log = logging.getLogger(__name__)


markdown_basic_jinja_template_txt = """
{% for s in lowstates %}
`{{s.id_full}}`
-----------------------------------------------------------------
 * state: {{s.state_function}}
 * name: `{{s.name}}`

{{s.markdown.requisites}}
{{s.markdown.details}}

{%- endfor %}
"""

markdown_default_jinja_template_txt = """
Configuration Managment
===============================================================================

```
####################################################
fqdn: {{grains.get('fqdn')}}
os: {{grains.get('os')}}
osfinger: {{grains.get('osfinger')}}
mem_total: {{grains.get('mem_total')}}MB
num_cpus: {{grains.get('num_cpus')}}
ipv4: {{grains.get('ipv4')}}
master: {{opts.get('master')}}
####################################################
```

This system is fully or partly managed using Salt.

The following sections are a rendered view of what the configuration management system
controlled on this system. Each item is handled in order from top to bottom unless some
requisites like `require` force other ordering.

""" + markdown_basic_jinja_template_txt


markdown_advanced_jinja_template_txt = markdown_default_jinja_template_txt + """

{% if vars.get('doc_other', True) -%}
Other information
=====================================================================================

```

salt grain: ip_interfaces
-----------------------------------------------------------------
{{grains['ip_interfaces']|dictsort}}


salt grain: hwaddr_interfaces
-----------------------------------------------------------------
{{grains['hwaddr_interfaces']|dictsort}}

{% if not grains['os'] == 'Windows' %}

{% if salt['cmd.has_exec']('ip') -%}
# ip address show
-----------------------------------------------------------------
{{salt['cmd.run']('ip address show | sed "/valid_lft/d"')}}


# ip route list table all
-----------------------------------------------------------------
{{salt['cmd.run']('ip route list table all')}}
{% endif %}

{% if salt['cmd.has_exec']('iptables') %}
{%- if salt['cmd.has_exec']('iptables-save') -%}
# iptables-save
-----------------------------------------------------------------
{{salt['cmd.run']("iptables --list > /dev/null; iptables-save | \grep -v -F '#' | sed '/^:/s@\[[0-9]\{1,\}:[0-9]\{1,\}\]@[0:0]@g'")}}


# ip6tables-save
-----------------------------------------------------------------
{{salt['cmd.run']("ip6tables --list > /dev/null; ip6tables-save | \grep -v -F '#' | sed '/^:/s@\[[0-9]\{1,\}:[0-9]\{1,\}\]@[0:0]@g'")}}
{%- else -%}
# iptables --list-rules
-----------------------------------------------------------------
{{salt['cmd.run']('iptables --list-rules')}}


# ip6tables --list-rules
-----------------------------------------------------------------
{{salt['cmd.run']('ip6tables --list-rules')}}
{% endif %}
{% endif %}

{% if salt['cmd.has_exec']('firewall-cmd') -%}
# firewall-cmd --list-all
-----------------------------------------------------------------
{{salt['cmd.run']('firewall-cmd --list-all')}}
{% endif %}

# mount
-----------------------------------------------------------------
{{salt['cmd.run']('mount')}}

{% endif %}
"""


def markdown_basic_jinja_template(**kwargs):
    '''
    Return text for a simple markdown jinja template

    This function can be used from the `highstate_doc.render` modules `jinja_template_function` option.
    '''
    return markdown_basic_jinja_template_txt


def markdown_default_jinja_template(**kwargs):
    '''
    Return text for a markdown jinja template that included a header

    This function can be used from the `highstate_doc.render` modules `jinja_template_function` option.
    '''
    return markdown_default_jinja_template_txt


def markdown_full_jinja_template(**kwargs):
    '''
    Return text for an advanced markdown jinja template

    This function can be used from the `highstate_doc.render` modules `jinja_template_function` option.
    '''
    return markdown_advanced_jinja_template_txt


def _get_config(**kwargs):
    '''
    Return configuration
    '''
    config = {
        'filter_id_regex': ['.*!doc_skip'],
        'filter_function_regex': [],
        'replace_text_regex': {},
        'proccesser': 'highstate_doc.proccesser_markdown',
        'max_render_file_size': 10000,
        'note': None
    }
    if '__salt__' in globals():
        config_key = '{0}.config'.format(__virtualname__)
        config.update(__salt__['config.get'](config_key, {}))
    # pylint: disable=C0201
    for k in set(config.keys()) & set(kwargs.keys()):
        config[k] = kwargs[k]
    return config


def read_file(name):
    '''
    output the contents of a file:

    this is a workaround if the cp.push module does not work.
    https://github.com/saltstack/salt/issues/37133

    help the master output the contents of a document
    that might be saved on the minions filesystem.

    .. code-block:: python

        #!/bin/python
        import os
        import salt.client
        s = salt.client.LocalClient()
        o = s.cmd('*', 'highstate_doc.read_file', ['/root/README.md'])
        for m in o:
            d = o.get(m)
            if d and not d.endswith('is not available.'):
                # mkdir m
                #directory = os.path.dirname(file_path)
                if not os.path.exists(m):
                    os.makedirs(m)
                with open(m + '/README.md','wb') as fin:
                    fin.write(d)
                print('ADDED: ' + m + '/README.md')
    '''
    out = ''
    try:
        with salt.utils.files.fopen(name, 'r') as f:
            out = salt.utils.stringutils.to_unicode(f.read())
    except Exception as ex:
        log.error(ex)
        return None
    return out


def render(jinja_template_text=None, jinja_template_function='highstate_doc.markdown_default_jinja_template', **kwargs):
    '''
    Render highstate to a text format (default Markdown)

    if `jinja_template_text` is not set, `jinja_template_function` is used.

    jinja_template_text: jinja text that the render uses to create the document.
    jinja_template_function: a salt module call that returns template text.
        options:
            highstate_doc.markdown_basic_jinja_template
            highstate_doc.markdown_default_jinja_template
            highstate_doc.markdown_full_jinja_template
    '''
    config = _get_config(**kwargs)
    lowstates = proccess_lowstates(**kwargs)
    # TODO: __env__,
    context = {
        'saltenv': None,
        'config': config,
        'lowstates': lowstates,
        'salt': __salt__,
        'pillar': __pillar__,
        'grains': __grains__,
        'opts': __opts__,
        'kwargs': kwargs,
    }
    template_text = jinja_template_text
    if template_text is None and jinja_template_function:
        template_text = __salt__[jinja_template_function](**kwargs)
    if template_text is None:
        raise Exception('No jinja template text')

    txt = tpl.render_jinja_tmpl(template_text, context, tmplpath=None)
    # after proccessing the template replace passwords or other data.
    rt = config.get('replace_text_regex')
    for r in rt:
        txt = re.sub(r, rt[r], txt)
    return txt


def _blacklist_filter(s, config):
    ss = s['state']
    sf = s['fun']
    state_function = '{0}.{1}'.format(s['state'], s['fun'])
    for b in config['filter_function_regex']:
        if re.match(b, state_function):
            return True
    for b in config['filter_id_regex']:
        if re.match(b, s['__id__']):
            return True
    return False


def proccess_lowstates(**kwargs):
    '''
    return proccessed lowstate data that was not blacklisted

    render_module_function is used to provide your own.
    defaults to from_lowstate
    '''
    states = []
    config = _get_config(**kwargs)
    proccesser = config.get('proccesser')
    ls = __salt__['state.show_lowstate']()

    if not isinstance(ls, list):
        raise Exception('ERROR: to see details run: [salt-call state.show_lowstate] <-----***-SEE-***')
    else:
        if len(ls) > 0:
            if not isinstance(ls[0], dict):
                raise Exception('ERROR: to see details run: [salt-call state.show_lowstate] <-----***-SEE-***')

    for s in ls:
        if _blacklist_filter(s, config):
            continue
        doc = __salt__[proccesser](s, config, **kwargs)
        states.append(doc)
    return states


def _state_data_to_yaml_string(data, whitelist=None, blacklist=None):
    '''
    return a data dict in yaml string format.
    '''
    y = {}
    if blacklist is None:
        # TODO: use salt defined STATE_REQUISITE_IN_KEYWORDS  STATE_RUNTIME_KEYWORDS  STATE_INTERNAL_KEYWORDS
        blacklist = ['__env__', '__id__', '__sls__', 'fun', 'name', 'context', 'order', 'state', 'require', 'require_in', 'watch', 'watch_in']
    kset = set(data.keys())
    if blacklist:
        kset -= set(blacklist)
    if whitelist:
        kset &= set(whitelist)
    for k in kset:
        y[k] = data[k]
    if len(y) == 0:
        return None
    return salt.utils.yaml.safe_dump(y, default_flow_style=False)


def _md_fix(text):
    '''
    sanitize text data that is to be displayed in a markdown code block
    '''
    return text.replace('```', '``[`][markdown parse fix]')


def _format_markdown_system_file(filename, config):
    ret = ''
    file_stats = __salt__['file.stats'](filename)
    y = _state_data_to_yaml_string(file_stats, whitelist=['user', 'group', 'mode', 'uid', 'gid', 'size'])
    if y:
        ret += 'file stat {1}\n```\n{0}```\n'.format(y, filename)
    file_size = file_stats.get('size')
    if file_size <= config.get('max_render_file_size'):
        is_binary = True
        try:
            # TODO: this is linux only should find somthing portable
            file_type = __salt__['cmd.shell']('\\file -i \'{0}\''.format(filename))
            if 'charset=binary' not in file_type:
                is_binary = False
        except Exception as ex:
            # likely on a windows system, set as not binary for now.
            is_binary = False
        if is_binary:
            file_data = '[[skipped binary data]]'
        else:
            with salt.utils.files.fopen(filename, 'r') as f:
                file_data = salt.utils.stringutils.to_unicode(f.read())
        file_data = _md_fix(file_data)
        ret += 'file data {1}\n```\n{0}\n```\n'.format(file_data, filename)
    else:
        ret += '```\n{0}\n```\n'.format('SKIPPED LARGE FILE!\nSet {0}:max_render_file_size > {1} to render.'.format('{0}.config'.format(__virtualname__), file_size))
    return ret


def _format_markdown_link(name):
    link = name
    symbals = '~`!@#$%^&*()+={}[]:;"<>,.?/|\'\\'
    for s in symbals:
        link = link.replace(s, '')
    link = link.replace(' ', '-')
    return link


def _format_markdown_requisite(state, stateid, makelink=True):
    '''
    format requisite as a link users can click
    '''
    fmt_id = '{0}: {1}'.format(state, stateid)
    if makelink:
        return ' * [{0}](#{1})\n'.format(fmt_id, _format_markdown_link(fmt_id))
    else:
        return ' * `{0}`\n'.format(fmt_id)


def proccesser_markdown(lowstate_item, config, **kwargs):
    '''
    Takes low state data and returns a dict of proccessed data
    that is by default used in a jinja template when rendering a markdown highstate_doc.

    This `lowstate_item_markdown` given a lowstate item, returns a dict like:

    .. code-block:: yaml

        vars:       # the raw lowstate_item that was proccessed
        id:         # the 'id' of the state.
        id_full:    # combo of the state type and id "state: id"
        state:      # name of the salt state module
        function:   # name of the state function
        name:       # value of 'name:' passed to the salt state module
        state_function:    # the state name and function name
        markdown:          # text data to describe a state
            requisites:    # requisite like [watch_in, require_in]
            details:       # state name, parameters and other details like file contents

    '''
    # TODO: switch or ... ext call.
    s = lowstate_item
    state_function = '{0}.{1}'.format(s['state'], s['fun'])
    id_full = '{0}: {1}'.format(s['state'], s['__id__'])

    # TODO: use salt defined STATE_REQUISITE_IN_KEYWORDS
    requisites = ''
    if s.get('watch'):
        requisites += 'run or update after changes in:\n'
        for w in s.get('watch', []):
            requisites += _format_markdown_requisite(w.items()[0][0], w.items()[0][1])
        requisites += '\n'
    if s.get('watch_in'):
        requisites += 'after changes, run or update:\n'
        for w in s.get('watch_in', []):
            requisites += _format_markdown_requisite(w.items()[0][0], w.items()[0][1])
        requisites += '\n'
    if s.get('require') and len(s.get('require')) > 0:
        requisites += 'require:\n'
        for w in s.get('require', []):
            requisites += _format_markdown_requisite(w.items()[0][0], w.items()[0][1])
        requisites += '\n'
    if s.get('require_in'):
        requisites += 'required in:\n'
        for w in s.get('require_in', []):
            requisites += _format_markdown_requisite(w.items()[0][0], w.items()[0][1])
        requisites += '\n'

    details = ''

    if state_function == 'highstate_doc.note':
        if 'contents' in s:
            details += '\n{0}\n'.format(s['contents'])
        if 'source' in s:
            text = __salt__['cp.get_file_str'](s['source'])
            if text:
                details += '\n{0}\n'.format(text)
            else:
                details += '\n{0}\n'.format('ERROR: opening {0}'.format(s['source']))

    if state_function == 'pkg.installed':
        pkgs = s.get('pkgs', s.get('name'))
        details += '\n```\ninstall: {0}\n```\n'.format(pkgs)

    if state_function == 'file.recurse':
        details += '''recurse copy of files\n'''
        y = _state_data_to_yaml_string(s)
        if y:
            details += '```\n{0}\n```\n'.format(y)
        if '!doc_recurse' in id_full:
            findfiles = __salt__['file.find'](path=s.get('name'), type='f')
            if len(findfiles) < 10 or '!doc_recurse_force' in id_full:
                for f in findfiles:
                    details += _format_markdown_system_file(f, config)
            else:
                details += ''' > Skipping because more than 10 files to display.\n'''
                details += ''' > HINT: to force include !doc_recurse_force in state id.\n'''
        else:
            details += ''' > For more details review logs and Salt state files.\n\n'''
            details += ''' > HINT: for improved docs use multiple file.managed states or file.archive, git.latest. etc.\n'''
            details += ''' > HINT: to force doc to show all files in path add !doc_recurse .\n'''

    if state_function == 'file.blockreplace':
        if s.get('content'):
            details += 'ensure block of content is in file\n```\n{0}\n```\n'.format(_md_fix(s['content']))
        if s.get('source'):
            text = '** source: ' + s.get('source')
            details += 'ensure block of content is in file\n```\n{0}\n```\n'.format(_md_fix(text))

    if state_function == 'file.managed':
        details += _format_markdown_system_file(s['name'], config)

    # if no state doc is created use default state as yaml
    if len(details) == 0:
        y = _state_data_to_yaml_string(s)
        if y:
            details += '```\n{0}```\n'.format(y)

    r = {
        'vars': lowstate_item,
        'state': s['state'],
        'name': s['name'],
        'function': s['fun'],
        'id': s['__id__'],
        'id_full': id_full,
        'state_function': state_function,
        'markdown': {
            'requisites': requisites.decode('utf-8'),
            'details': details.decode('utf-8')
        }
    }
    return r



# -*- coding: utf-8 -*-
'''
Return salt data via slack

..  versionadded:: 2015.5.0

The following fields can be set in the minion conf file:

.. code-block:: yaml

    slack.channel (required)
    slack.api_key (required)
    slack.username (required)
    slack.as_user (required to see the profile picture of your bot)
    slack.profile (optional)
    slack.changes(optional, only show changes and failed states)
    slack.yaml_format(optional, format the json in yaml format)


Alternative configuration values can be used by prefacing the configuration.
Any values not found in the alternative configuration will be pulled from
the default location:

.. code-block:: yaml

    slack.channel
    slack.api_key
    slack.username
    slack.as_user

Slack settings may also be configured as:

.. code-block:: yaml

    slack:
        channel: RoomName
        api_key: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        username: user
        as_user: true

    alternative.slack:
        room_id: RoomName
        api_key: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        from_name: user@email.com

    slack_profile:
        slack.api_key: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        slack.from_name: user@email.com

    slack:
        profile: slack_profile
        channel: RoomName

    alternative.slack:
        profile: slack_profile
        channel: RoomName

To use the Slack returner, append '--return slack' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return slack

To use the alternative configuration, append '--return_config alternative' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return slack --return_config alternative

To override individual configuration items, append --return_kwargs '{"key:": "value"}' to the salt command.

.. versionadded:: 2016.3.0

.. code-block:: bash

    salt '*' test.ping --return slack --return_kwargs '{"channel": "#random"}'

'''
from __future__ import absolute_import, print_function, unicode_literals

# Import Python libs
import pprint
import logging

# pylint: disable=import-error,no-name-in-module,redefined-builtin
import salt.ext.six.moves.http_client
from salt.ext.six.moves.urllib.parse import urlencode as _urlencode
# pylint: enable=import-error,no-name-in-module,redefined-builtin

# Import Salt Libs
import salt.returners
import salt.utils.slack
import salt.utils.yaml

log = logging.getLogger(__name__)

__virtualname__ = 'slack'


def _get_options(ret=None):
    '''
    Get the slack options from salt.
    '''

    defaults = {'channel': '#general'}

    attrs = {'slack_profile': 'profile',
             'channel': 'channel',
             'username': 'username',
             'as_user': 'as_user',
             'api_key': 'api_key',
             'changes': 'changes',
             'yaml_format': 'yaml_format',
             }

    profile_attr = 'slack_profile'

    profile_attrs = {'from_jid': 'from_jid',
                     'api_key': 'api_key',
                     'api_version': 'api_key'
                     }

    _options = salt.returners.get_returner_options(__virtualname__,
                                                   ret,
                                                   attrs,
                                                   profile_attr=profile_attr,
                                                   profile_attrs=profile_attrs,
                                                   __salt__=__salt__,
                                                   __opts__=__opts__,
                                                   defaults=defaults)
    return _options


def __virtual__():
    '''
    Return virtual name of the module.

    :return: The virtual name of the module.
    '''
    return __virtualname__


def _post_message(channel,
                  message,
                  username,
                  as_user,
                  api_key=None):
    '''
    Send a message to a Slack room.
    :param channel:     The room name.
    :param message:     The message to send to the Slack room.
    :param username:    Specify who the message is from.
    :param as_user:     Sets the profile picture which have been added through Slack itself.
    :param api_key:     The Slack api key, if not specified in the configuration.
    :param api_version: The Slack api version, if not specified in the configuration.
    :return:            Boolean if message was sent successfully.
    '''

    parameters = dict()
    parameters['channel'] = channel
    parameters['username'] = username
    parameters['as_user'] = as_user
    parameters['text'] = '```' + message + '```'  # pre-formatted, fixed-width text

    # Slack wants the body on POST to be urlencoded.
    result = salt.utils.slack.query(function='message',
                                    api_key=api_key,
                                    method='POST',
                                    header_dict={'Content-Type': 'application/x-www-form-urlencoded'},
                                    data=_urlencode(parameters))

    log.debug('Slack message post result: %s', result)
    if result:
        return True
    else:
        return False


def returner(ret):
    '''
    Send an slack message with the data
    '''

    _options = _get_options(ret)

    channel = _options.get('channel')
    username = _options.get('username')
    as_user = _options.get('as_user')
    api_key = _options.get('api_key')
    changes = _options.get('changes')
    yaml_format = _options.get('yaml_format')

    if not channel:
        log.error('slack.channel not defined in salt config')
        return

    if not username:
        log.error('slack.username not defined in salt config')
        return

    if not as_user:
        log.error('slack.as_user not defined in salt config')
        return

    if not api_key:
        log.error('slack.api_key not defined in salt config')
        return

    returns = ret.get('return')
    if changes is True:
        returns = dict((key, value) for key, value in returns.items() if value['result'] is not True or value['changes'])

    if yaml_format is True:
        returns = salt.utils.yaml.safe_dump(returns)
    else:
        returns = pprint.pformat(returns)

    message = ('id: {0}\r\n'
               'function: {1}\r\n'
               'function args: {2}\r\n'
               'jid: {3}\r\n'
               'return: {4}\r\n').format(
                    ret.get('id'),
                    ret.get('fun'),
                    ret.get('fun_args'),
                    ret.get('jid'),
                    returns)

    slack = _post_message(channel,
                          message,
                          username,
                          as_user,
                          api_key)
    return slack



# -*- coding: utf-8 -*-
'''
Network Config
==============

Manage the configuration on a network device given a specific static config or template.

:codeauthor: Mircea Ulinic <mircea@cloudflare.com> & Jerome Fleury <jf@cloudflare.com>
:maturity:   new
:depends:    napalm
:platform:   unix

Dependencies
------------
- :mod:`NAPALM proxy minion <salt.proxy.napalm>`
- :mod:`Network-related basic features execution module <salt.modules.napalm_network>`

.. versionadded:: 2017.7.0
'''

from __future__ import absolute_import, print_function, unicode_literals

import logging
log = logging.getLogger(__name__)

# import NAPALM utils
import salt.utils.napalm
import salt.utils.versions

# ----------------------------------------------------------------------------------------------------------------------
# state properties
# ----------------------------------------------------------------------------------------------------------------------

__virtualname__ = 'netconfig'

# ----------------------------------------------------------------------------------------------------------------------
# global variables
# ----------------------------------------------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------------------------------------------
# property functions
# ----------------------------------------------------------------------------------------------------------------------


def __virtual__():
    '''
    NAPALM library must be installed for this module to work and run in a (proxy) minion.
    '''
    return salt.utils.napalm.virtual(__opts__, __virtualname__, __file__)

# ----------------------------------------------------------------------------------------------------------------------
# helper functions -- will not be exported
# ----------------------------------------------------------------------------------------------------------------------


def _update_config(template_name,
                   template_source=None,
                   template_path=None,
                   template_hash=None,
                   template_hash_name=None,
                   template_user='root',
                   template_group='root',
                   template_mode='755',
                   template_attrs='--------------e----',
                   saltenv=None,
                   template_engine='jinja',
                   skip_verify=False,
                   defaults=None,
                   test=False,
                   commit=True,
                   debug=False,
                   replace=False,
                   **template_vars):
    '''
    Call the necessary functions in order to execute the state.
    For the moment this only calls the ``net.load_template`` function from the
    :mod:`Network-related basic features execution module <salt.modules.napalm_network>`, but this may change in time.
    '''

    return __salt__['net.load_template'](template_name,
                                         template_source=template_source,
                                         template_path=template_path,
                                         template_hash=template_hash,
                                         template_hash_name=template_hash_name,
                                         template_user=template_user,
                                         template_group=template_group,
                                         template_mode=template_mode,
                                         template_attrs=template_attrs,
                                         saltenv=saltenv,
                                         template_engine=template_engine,
                                         skip_verify=skip_verify,
                                         defaults=defaults,
                                         test=test,
                                         commit=commit,
                                         debug=debug,
                                         replace=replace,
                                         **template_vars)

# ----------------------------------------------------------------------------------------------------------------------
# callable functions
# ----------------------------------------------------------------------------------------------------------------------


def managed(name,
            template_name,
            template_source=None,
            template_path=None,
            template_hash=None,
            template_hash_name=None,
            template_user='root',
            template_group='root',
            template_mode='755',
            template_attrs='--------------e----',
            saltenv=None,
            template_engine='jinja',
            skip_verify=False,
            defaults=None,
            test=False,
            commit=True,
            debug=False,
            replace=False,
            **template_vars):

    '''
    Manages the configuration on network devices.

    By default this state will commit the changes on the device. If there are no changes required, it does not commit
    and the field ``already_configured`` from the output dictionary will be set as ``True`` to notify that.

    To avoid committing the configuration, set the argument ``test`` to ``True`` (or via the CLI argument ``test=True``)
    and will discard (dry run).

    To preserve the changes, set ``commit`` to ``False`` (either as CLI argument, either as state parameter).
    However, this is recommended to be used only in exceptional cases when there are applied few consecutive states
    and/or configuration changes. Otherwise the user might forget that the config DB is locked and the candidate config
    buffer is not cleared/merged in the running config.

    To replace the config, set ``replace`` to ``True``. This option is recommended to be used with caution!

    .. warning::
        The support for NAPALM native templates will be dropped beginning with Salt Fluorine.
        Implicitly, the ``template_path`` argument will be deprecated and removed.

    template_name
        Identifies path to the template source. The template can be either stored on the local machine,
        either remotely.
        The recommended location is under the ``file_roots`` as specified in the master config file.
        For example, let's suppose the ``file_roots`` is configured as:

        .. code-block:: yaml

            file_roots:
                base:
                    - /etc/salt/states

        Placing the template under ``/etc/salt/states/templates/example.jinja``, it can be used as
        ``salt://templates/example.jinja``.
        Alternatively, for local files, the user can specify the absolute path.
        If remotely, the source can be retrieved via ``http``, ``https`` or ``ftp``.

        Examples:

        - ``salt://my_template.jinja``
        - ``/absolute/path/to/my_template.jinja``
        - ``http://example.com/template.cheetah``
        - ``https:/example.com/template.mako``
        - ``ftp://example.com/template.py``

    template_source: None
        Inline config template to be rendered and loaded on the device.

    template_path: None
        Required only in case the argument ``template_name`` provides only the file basename.
        E.g.: if ``template_name`` is specified as ``my_template.jinja``, in order to find the
        template, this argument must be provided: ``template_path: /absolute/path/to/``.

    template_hash: None
        Hash of the template file. Format: ``{hash_type: 'md5', 'hsum': <md5sum>}``

    template_hash_name: None
        When ``template_hash`` refers to a remote file, this specifies the filename to look for in that file.

    template_group: root
        Owner of file.

    template_user: root
        Group owner of file.

    template_mode: 755
        Permissions of file

    template_attrs: "--------------e----"
        Attributes of file (see `man lsattr`)

        .. versionadded:: 2018.3.0

    saltenv: base
        Specifies the template environment. This will influence the relative imports inside the templates.

    template_engine: jinja
        The following templates engines are supported:

        - :mod:`cheetah<salt.renderers.cheetah>`
        - :mod:`genshi<salt.renderers.genshi>`
        - :mod:`jinja<salt.renderers.jinja>`
        - :mod:`mako<salt.renderers.mako>`
        - :mod:`py<salt.renderers.py>`
        - :mod:`wempy<salt.renderers.wempy>`

    skip_verify: False
        If ``True``, hash verification of remote file sources (``http://``, ``https://``, ``ftp://``) will be skipped,
        and the ``source_hash`` argument will be ignored.

        .. versionchanged:: 2017.7.1

    test: False
        Dry run? If set to ``True``, will apply the config, discard and return the changes. Default: ``False``
        (will commit the changes on the device).

    commit: True
        Commit? Default: ``True``.

    debug: False
        Debug mode. Will insert a new key under the output dictionary, as ``loaded_config`` containing the raw
        result after the template was rendered.

        .. note::
            This argument cannot be used directly on the command line. Instead,
            it can be passed through the ``pillar`` variable when executing
            either of the :py:func:`state.sls <salt.modules.state.sls>` or
            :py:func:`state.apply <salt.modules.state.apply>` (see below for an
            example).

    replace: False
        Load and replace the configuration. Default: ``False`` (will apply load merge).

    defaults: None
        Default variables/context passed to the template.

    template_vars
        Dictionary with the arguments/context to be used when the template is rendered. Do not explicitly specify this
        argument. This represents any other variable that will be sent to the template rendering system. Please
        see an example below! In both ``ntp_peers_example_using_pillar`` and ``ntp_peers_example``, ``peers`` is sent as
        template variable.

    SLS Example (e.g.: under salt://router/config.sls) :

    .. code-block:: yaml

        whole_config_example:
            netconfig.managed:
                - template_name: salt://path/to/complete_config.jinja
                - debug: True
                - replace: True
        bgp_config_example:
            netconfig.managed:
                - template_name: /absolute/path/to/bgp_neighbors.mako
                - template_engine: mako
        prefix_lists_example:
            netconfig.managed:
                - template_name: prefix_lists.cheetah
                - template_path: /absolute/path/to/
                - debug: True
                - template_engine: cheetah
        ntp_peers_example:
            netconfig.managed:
                - template_name: http://bit.ly/2gKOj20
                - skip_verify: False
                - debug: True
                - peers:
                    - 192.168.0.1
                    - 192.168.0.1
        ntp_peers_example_using_pillar:
            netconfig.managed:
                - template_name: http://bit.ly/2gKOj20
                - peers: {{ pillar.get('ntp.peers', []) }}

    Usage examples:

    .. code-block:: bash

        $ sudo salt 'juniper.device' state.sls router.config test=True

        $ sudo salt -N all-routers state.sls router.config pillar="{'debug': True}"

    ``router.config`` depends on the location of the SLS file (see above). Running this command, will be executed all
    five steps from above. These examples above are not meant to be used in a production environment, their sole purpose
    is to provide usage examples.

    Output example:

    .. code-block:: bash

        $ sudo salt 'juniper.device' state.sls router.config test=True
        juniper.device:
        ----------
                  ID: ntp_peers_example_using_pillar
            Function: netconfig.managed
              Result: None
             Comment: Testing mode: Configuration discarded.
             Started: 12:01:40.744535
            Duration: 8755.788 ms
             Changes:
                      ----------
                      diff:
                          [edit system ntp]
                               peer 192.168.0.1 { ... }
                          +    peer 172.17.17.1;
                          +    peer 172.17.17.3;

        Summary for juniper.device
        ------------
        Succeeded: 1 (unchanged=1, changed=1)
        Failed:    0
        ------------
        Total states run:     1
        Total run time:   8.756 s

    Raw output example (useful when the output is reused in other states/execution modules):

    .. code-block:: bash

        $ sudo salt --out=pprint 'juniper.device' state.sls router.config test=True debug=True

    .. code-block:: python

        {
            'juniper.device': {
                'netconfig_|-ntp_peers_example_using_pillar_|-ntp_peers_example_using_pillar_|-managed': {
                    '__id__': 'ntp_peers_example_using_pillar',
                    '__run_num__': 0,
                    'already_configured': False,
                    'changes': {
                        'diff': '[edit system ntp]   peer 192.168.0.1 { ... }+   peer 172.17.17.1;+   peer 172.17.17.3;'
                    },
                    'comment': 'Testing mode: Configuration discarded.',
                    'duration': 7400.759,
                    'loaded_config': 'system {  ntp {  peer 172.17.17.1;  peer 172.17.17.3; } }',
                    'name': 'ntp_peers_example_using_pillar',
                    'result': None,
                    'start_time': '12:09:09.811445'
                }
            }
        }
    '''
    if template_path:
        salt.utils.versions.warn_until(
            'Fluorine',
            'Use of `template_path` detected. This argument will be removed in Salt Fluorine.'
        )
    ret = salt.utils.napalm.default_ret(name)

    # the user can override the flags the equivalent CLI args
    # which have higher precedence
    test = __salt__['config.merge']('test', test)
    debug = __salt__['config.merge']('debug', debug)
    commit = __salt__['config.merge']('commit', commit)
    replace = __salt__['config.merge']('replace', replace)  # this might be a bit risky
    skip_verify = __salt__['config.merge']('skip_verify', skip_verify)

    config_update_ret = _update_config(template_name,
                                       template_source=template_source,
                                       template_path=template_path,
                                       template_hash=template_hash,
                                       template_hash_name=template_hash_name,
                                       template_user=template_user,
                                       template_group=template_group,
                                       template_mode=template_mode,
                                       template_attrs=template_attrs,
                                       saltenv=saltenv,
                                       template_engine=template_engine,
                                       skip_verify=skip_verify,
                                       defaults=defaults,
                                       test=test,
                                       commit=commit,
                                       debug=debug,
                                       replace=replace,
                                       **template_vars)

    return salt.utils.napalm.loaded_ret(ret, config_update_ret, test, debug)



# -*- coding: utf-8 -*-
'''
Openstack Cloud Driver
======================

:depends: `shade <https://pypi.python.org/pypi/shade>`_

OpenStack is an open source project that is in use by a number a cloud
providers, each of which have their own ways of using it.

This OpenStack driver uses a the shade python module which is managed by the
OpenStack Infra team.  This module is written to handle all the different
versions of different OpenStack tools for salt, so most commands are just passed
over to the module to handle everything.

Provider
--------

There are two ways to configure providers for this driver.  The first one is to
just let shade handle everything, and configure using os-client-config_ and
setting up `/etc/openstack/clouds.yml`.

.. code-block:: yaml

    clouds:
      democloud:
        region_name: RegionOne
        auth:
          username: 'demo'
          password: secret
          project_name: 'demo'
          auth_url: 'http://openstack/identity'

And then this can be referenced in the salt provider based on the `democloud`
name.

.. code-block:: yaml

    myopenstack:
      driver: openstack
      cloud: democloud
      region_name: RegionOne

This allows for just using one configuration for salt-cloud and for any other
openstack tools which are all using `/etc/openstack/clouds.yml`

The other method allows for specifying everything in the provider config,
instead of using the extra configuration file.  This will allow for passing
salt-cloud configs only through pillars for minions without having to write a
clouds.yml file on each minion.abs

.. code-block:: yaml

    myopenstack:
      driver: openstack
      region_name: RegionOne
      auth:
        username: 'demo'
        password: secret
        project_name: 'demo'
        auth_url: 'http://openstack/identity'

Or if you need to use a profile to setup some extra stuff, it can be passed as a
`profile` to use any of the vendor_ config options.

.. code-block:: yaml

    myrackspace:
      driver: openstack
      profile: rackspace
      auth:
        username: rackusername
        api_key: myapikey
      region_name: ORD
      auth_type: rackspace_apikey

And this will pull in the profile for rackspace and setup all the correct
options for the auth_url and different api versions for services.


Profile
-------

Most of the options for building servers are just passed on to the
create_server_ function from shade.

The salt specific ones are:

  - ssh_key_file: The path to the ssh key that should be used to login to the machine to bootstrap it
  - ssh_key_file: The name of the keypair in openstack
  - userdata_template: The renderer to use if the userdata is a file that is templated. Default: False
  - ssh_interface: The interface to use to login for bootstrapping: public_ips, private_ips, floating_ips, fixed_ips

.. code-block:: yaml

    centos:
      provider: myopenstack
      image: CentOS 7
      size: ds1G
      ssh_key_name: mykey
      ssh_key_file: /root/.ssh/id_rsa

This is the minimum setup required.

If metadata is set to make sure that the host has finished setting up the
`wait_for_metadata` can be set.

.. code-block:: yaml

    centos:
      provider: myopenstack
      image: CentOS 7
      size: ds1G
      ssh_key_name: mykey
      ssh_key_file: /root/.ssh/id_rsa
      meta:
        build_config: rack_user_only
      wait_for_metadata:
        rax_service_level_automation: Complete
        rackconnect_automation_status: DEPLOYED

Anything else from the create_server_ docs can be passed through here.

- **image**: Image dict, name or ID to boot with. image is required
            unless boot_volume is given.
- **flavor**: Flavor dict, name or ID to boot onto.
- **auto_ip**: Whether to take actions to find a routable IP for
              the server. (defaults to True)
- **ips**: List of IPs to attach to the server (defaults to None)
- **ip_pool**: Name of the network or floating IP pool to get an
              address from. (defaults to None)
- **root_volume**: Name or ID of a volume to boot from
                  (defaults to None - deprecated, use boot_volume)
- **boot_volume**: Name or ID of a volume to boot from
                  (defaults to None)
- **terminate_volume**: If booting from a volume, whether it should
                       be deleted when the server is destroyed.
                       (defaults to False)
- **volumes**: (optional) A list of volumes to attach to the server
- **meta**: (optional) A dict of arbitrary key/value metadata to
           store for this server. Both keys and values must be
           <=255 characters.
- **files**: (optional, deprecated) A dict of files to overwrite
            on the server upon boot. Keys are file names (i.e.
            ``/etc/passwd``) and values
            are the file contents (either as a string or as a
            file-like object). A maximum of five entries is allowed,
            and each file must be 10k or less.
- **reservation_id**: a UUID for the set of servers being requested.
- **min_count**: (optional extension) The minimum number of
                servers to launch.
- **max_count**: (optional extension) The maximum number of
                servers to launch.
- **security_groups**: A list of security group names
- **userdata**: user data to pass to be exposed by the metadata
            server this can be a file type object as well or a
            string.
- **key_name**: (optional extension) name of previously created
            keypair to inject into the instance.
- **availability_zone**: Name of the availability zone for instance
                        placement.
- **block_device_mapping**: (optional) A dict of block
            device mappings for this server.
- **block_device_mapping_v2**: (optional) A dict of block
            device mappings for this server.
- **nics**:  (optional extension) an ordered list of nics to be
            added to this server, with information about
            connected networks, fixed IPs, port etc.
- **scheduler_hints**: (optional extension) arbitrary key-value pairs
                  specified by the client to help boot an instance
- **config_drive**: (optional extension) value for config drive
                  either boolean, or volume-id
- **disk_config**: (optional extension) control how the disk is
                  partitioned when the server is created.  possible
                  values are 'AUTO' or 'MANUAL'.
- **admin_pass**: (optional extension) add a user supplied admin
                 password.
- **timeout**: (optional) Seconds to wait, defaults to 60.
              See the ``wait`` parameter.
- **reuse_ips**: (optional) Whether to attempt to reuse pre-existing
                           floating ips should a floating IP be
                           needed (defaults to True)
- **network**: (optional) Network dict or name or ID to attach the
              server to.  Mutually exclusive with the nics parameter.
              Can also be be a list of network names or IDs or
              network dicts.
- **boot_from_volume**: Whether to boot from volume. 'boot_volume'
                       implies True, but boot_from_volume=True with
                       no boot_volume is valid and will create a
                       volume from the image and use that.
- **volume_size**: When booting an image from volume, how big should
                  the created volume be? Defaults to 50.
- **nat_destination**: Which network should a created floating IP
                      be attached to, if it's not possible to
                      infer from the cloud's configuration.
                      (Optional, defaults to None)
- **group**: ServerGroup dict, name or id to boot the server in.
            If a group is provided in both scheduler_hints and in
            the group param, the group param will win.
            (Optional, defaults to None)

.. note::

    If there is anything added, that is not in this list, it can be added to an `extras`
    dictionary for the profile, and that will be to the create_server function.

.. _create_server: https://docs.openstack.org/shade/latest/user/usage.html#shade.OpenStackCloud.create_server
.. _vendor: https://docs.openstack.org/os-client-config/latest/user/vendor-support.html
.. _os-client-config: https://docs.openstack.org/os-client-config/latest/user/configuration.html#config-files
'''
from __future__ import absolute_import, print_function, unicode_literals

# Import Python Libs
import copy
import logging
import os
import pprint
import socket

# Import Salt Libs
import salt.utils.json
import salt.config as config
from salt.ext import six
from salt.exceptions import (
    SaltCloudSystemExit,
    SaltCloudExecutionTimeout,
    SaltCloudExecutionFailure,
    SaltCloudConfigError,
)

# Import 3rd-Party Libs
try:
    import shade.openstackcloud
    import shade.exc
    import os_client_config
    HAS_SHADE = True
except ImportError:
    HAS_SHADE = False

log = logging.getLogger(__name__)
__virtualname__ = 'openstack'


def __virtual__():
    '''
    Check for OpenStack dependencies
    '''
    if get_configured_provider() is False:
        return False
    if get_dependencies() is False:
        return False
    return __virtualname__


def get_configured_provider():
    '''
    Return the first configured instance.
    '''
    return config.is_provider_configured(
        __opts__, __active_provider_name__ or __virtualname__,
        ('auth', 'region_name'), log_message=False,
    ) or config.is_provider_configured(
        __opts__, __active_provider_name__ or __virtualname__,
        ('cloud', 'region_name')
    )


def get_dependencies():
    '''
    Warn if dependencies aren't met.
    '''
    deps = {
        'shade': HAS_SHADE,
        'os_client_config': HAS_SHADE,
    }
    return config.check_driver_dependencies(
        __virtualname__,
        deps
    )


def preferred_ip(vm_, ips):
    '''
    Return the preferred Internet protocol. Either 'ipv4' (default) or 'ipv6'.
    '''
    proto = config.get_cloud_config_value(
        'protocol', vm_, __opts__, default='ipv4', search_global=False
    )

    family = socket.AF_INET
    if proto == 'ipv6':
        family = socket.AF_INET6
    for ip in ips:
        try:
            socket.inet_pton(family, ip)
            return ip
        except Exception:
            continue
    return False


def ssh_interface(vm_):
    '''
    Return the ssh_interface type to connect to. Either 'public_ips' (default)
    or 'private_ips'.
    '''
    return config.get_cloud_config_value('ssh_interface', vm_, __opts__, default='public_ips', search_global=False)


def get_conn():
    '''
    Return a conn object for the passed VM data
    '''
    if __active_provider_name__ in __context__:
        return __context__[__active_provider_name__]
    vm_ = get_configured_provider()
    profile = vm_.pop('profile', None)
    if profile is not None:
        vm_ = __utils__['dictupdate.update'](os_client_config.vendors.get_profile(profile), vm_)
    conn = shade.openstackcloud.OpenStackCloud(cloud_config=None, **vm_)
    if __active_provider_name__ is not None:
        __context__[__active_provider_name__] = conn
    return conn


def list_nodes(conn=None, call=None):
    '''
    Return a list of VMs

    CLI Example

    .. code-block:: bash

        salt-cloud -f list_nodes myopenstack

    '''
    if call == 'action':
        raise SaltCloudSystemExit(
            'The list_nodes function must be called with -f or --function.'
        )
    ret = {}
    for node, info in list_nodes_full(conn=conn).items():
        for key in ('id', 'name', 'size', 'state', 'private_ips', 'public_ips', 'floating_ips', 'fixed_ips', 'image'):
            ret.setdefault(node, {}).setdefault(key, info.get(key))

    return ret


def list_nodes_min(conn=None, call=None):
    '''
    Return a list of VMs with minimal information

    CLI Example

    .. code-block:: bash

        salt-cloud -f list_nodes_min myopenstack

    '''
    if call == 'action':
        raise SaltCloudSystemExit(
            'The list_nodes_min function must be called with -f or --function.'
        )
    if conn is None:
        conn = get_conn()
    ret = {}
    for node in conn.list_servers(bare=True):
        ret[node.name] = {'id': node.id, 'state': node.status}
    return ret


def _get_ips(node, addr_type='public'):
    ret = []
    for _, interface in node.addresses.items():
        for addr in interface:
            if addr_type in ('floating', 'fixed') and addr_type == addr.get('OS-EXT-IPS:type'):
                ret.append(addr['addr'])
            elif addr_type == 'public' and __utils__['cloud.is_public_ip'](addr['addr']):
                ret.append(addr['addr'])
            elif addr_type == 'private' and not __utils__['cloud.is_public_ip'](addr['addr']):
                ret.append(addr['addr'])
    return ret


def list_nodes_full(conn=None, call=None):
    '''
    Return a list of VMs with all the information about them

    CLI Example

    .. code-block:: bash

        salt-cloud -f list_nodes_full myopenstack

    '''
    if call == 'action':
        raise SaltCloudSystemExit(
            'The list_nodes_full function must be called with -f or --function.'
        )
    if conn is None:
        conn = get_conn()
    ret = {}
    for node in conn.list_servers(detailed=True):
        ret[node.name] = dict(node)
        ret[node.name]['id'] = node.id
        ret[node.name]['name'] = node.name
        ret[node.name]['size'] = node.flavor.name
        ret[node.name]['state'] = node.status
        ret[node.name]['private_ips'] = _get_ips(node, 'private')
        ret[node.name]['public_ips'] = _get_ips(node, 'public')
        ret[node.name]['floating_ips'] = _get_ips(node, 'floating')
        ret[node.name]['fixed_ips'] = _get_ips(node, 'fixed')
        ret[node.name]['image'] = node.image.name
    return ret


def list_nodes_select(conn=None, call=None):
    '''
    Return a list of VMs with the fields from `query.selection`

    CLI Example

    .. code-block:: bash

        salt-cloud -f list_nodes_full myopenstack

    '''
    if call == 'action':
        raise SaltCloudSystemExit(
            'The list_nodes_select function must be called with -f or --function.'
        )
    return __utils__['cloud.list_nodes_select'](
        list_nodes(conn, 'function'), __opts__['query.selection'], call,
    )


def show_instance(name, conn=None, call=None):
    '''
    Get VM on this OpenStack account

    name

        name of the instance

    CLI Example

    .. code-block:: bash

        salt-cloud -a show_instance myserver

    '''
    if call != 'action':
        raise SaltCloudSystemExit(
            'The show_instance action must be called with -a or --action.'
        )
    if conn is None:
        conn = get_conn()

    node = conn.get_server(name, bare=True)
    ret = dict(node)
    ret['id'] = node.id
    ret['name'] = node.name
    ret['size'] = conn.get_flavor(node.flavor.id).name
    ret['state'] = node.status
    ret['private_ips'] = _get_ips(node, 'private')
    ret['public_ips'] = _get_ips(node, 'public')
    ret['floating_ips'] = _get_ips(node, 'floating')
    ret['fixed_ips'] = _get_ips(node, 'fixed')
    if isinstance(node.image, six.string_types):
        ret['image'] = node.image
    else:
        ret['image'] = conn.get_image(node.image.id).name
    return ret


def avail_images(conn=None, call=None):
    '''
    List available images for OpenStack

    CLI Example

    .. code-block:: bash

        salt-cloud -f avail_images myopenstack
        salt-cloud --list-images myopenstack

    '''
    if call == 'action':
        raise SaltCloudSystemExit(
            'The avail_images function must be called with '
            '-f or --function, or with the --list-images option'
        )
    if conn is None:
        conn = get_conn()
    return conn.list_images()


def avail_sizes(conn=None, call=None):
    '''
    List available sizes for OpenStack

    CLI Example

    .. code-block:: bash

        salt-cloud -f avail_sizes myopenstack
        salt-cloud --list-sizes myopenstack

    '''
    if call == 'action':
        raise SaltCloudSystemExit(
            'The avail_sizes function must be called with '
            '-f or --function, or with the --list-sizes option'
        )
    if conn is None:
        conn = get_conn()
    return conn.list_flavors()


def list_networks(conn=None, call=None):
    '''
    List networks for OpenStack

    CLI Example

    .. code-block:: bash

        salt-cloud -f list_networks myopenstack

    '''
    if call == 'action':
        raise SaltCloudSystemExit(
            'The list_networks function must be called with '
            '-f or --function'
        )
    if conn is None:
        conn = get_conn()
    return conn.list_networks()


def list_subnets(conn=None, call=None, kwargs=None):
    '''
    List subnets in a virtual network

    network
        network to list subnets of

    .. code-block:: bash

        salt-cloud -f list_subnets myopenstack network=salt-net

    '''
    if call == 'action':
        raise SaltCloudSystemExit(
            'The list_subnets function must be called with '
            '-f or --function.'
        )
    if conn is None:
        conn = get_conn()
    if kwargs is None or (isinstance(kwargs, dict) and 'network' not in kwargs):
        raise SaltCloudSystemExit(
            'A `network` must be specified'
        )
    return conn.list_subnets(filters={'network': kwargs['network']})


def _clean_create_kwargs(**kwargs):
    '''
    Sanatize kwargs to be sent to create_server
    '''
    VALID_OPTS = {
        'name': six.string_types,
        'image': six.string_types,
        'flavor': six.string_types,
        'auto_ip': bool,
        'ips': list,
        'ip_pool': six.string_types,
        'root_volume': six.string_types,
        'boot_volume': six.string_types,
        'terminate_volume': bool,
        'volumes': list,
        'meta': dict,
        'files': dict,
        'reservation_id': six.string_types,
        'security_groups': list,
        'key_name': six.string_types,
        'availability_zone': six.string_types,
        'block_device_mapping': dict,
        'block_device_mapping_v2': dict,
        'nics': list,
        'scheduler_hints': dict,
        'config_drive': bool,
        'disk_config': six.string_types,  # AUTO or MANUAL
        'admin_pass': six.string_types,
        'wait': bool,
        'timeout': int,
        'reuse_ips': bool,
        'network': dict,
        'boot_from_volume': bool,
        'volume_size': int,
        'nat_destination': six.string_types,
        'group': six.string_types,
        'userdata': six.string_types,
    }
    extra = kwargs.pop('extra', {})
    for key, value in six.iteritems(kwargs.copy()):
        if key in VALID_OPTS:
            if isinstance(value, VALID_OPTS[key]):
                continue
            log.error('Error %s: %s is not of type %s', key, value, VALID_OPTS[key])
        kwargs.pop(key)
    return __utils__['dictupdate.update'](kwargs, extra)


def request_instance(vm_, conn=None, call=None):
    '''
    Request an instance to be built
    '''
    if call == 'function':
        # Technically this function may be called other ways too, but it
        # definitely cannot be called with --function.
        raise SaltCloudSystemExit(
            'The request_instance action must be called with -a or --action.'
        )
    kwargs = copy.deepcopy(vm_)
    log.info('Creating Cloud VM %s', vm_['name'])
    __utils__['cloud.check_name'](vm_['name'], 'a-zA-Z0-9._-')
    conn = get_conn()
    userdata = config.get_cloud_config_value(
        'userdata', vm_, __opts__, search_global=False, default=None
    )
    if userdata is not None and os.path.isfile(userdata):
        try:
            with __utils__['files.fopen'](userdata, 'r') as fp_:
                kwargs['userdata'] = __utils__['cloud.userdata_template'](
                    __opts__, vm_, fp_.read()
                )
        except Exception as exc:
            log.exception(
                'Failed to read userdata from %s: %s', userdata, exc)
    if 'size' in kwargs:
        kwargs['flavor'] = kwargs.pop('size')
    kwargs['key_name'] = config.get_cloud_config_value(
        'ssh_key_name', vm_, __opts__, search_global=False, default=None
    )
    kwargs['wait'] = True
    try:
        conn.create_server(**_clean_create_kwargs(**kwargs))
    except shade.exc.OpenStackCloudException as exc:
        log.error('Error creating server %s: %s', vm_['name'], exc)
        destroy(vm_['name'], conn=conn, call='action')
        raise SaltCloudSystemExit(six.text_type(exc))

    return show_instance(vm_['name'], conn=conn, call='action')


def create(vm_):
    '''
    Create a single VM from a data dict
    '''
    deploy = config.get_cloud_config_value('deploy', vm_, __opts__)
    key_filename = config.get_cloud_config_value(
        'ssh_key_file', vm_, __opts__, search_global=False, default=None
    )
    if key_filename is not None and not os.path.isfile(key_filename):
        raise SaltCloudConfigError(
            'The defined ssh_key_file \'{0}\' does not exist'.format(
                key_filename
            )
        )

    vm_['key_filename'] = key_filename

    __utils__['cloud.fire_event'](
        'event',
        'starting create',
        'salt/cloud/{0}/creating'.format(vm_['name']),
        args=__utils__['cloud.filter_event']('creating', vm_, ['name', 'profile', 'provider', 'driver']),
        sock_dir=__opts__['sock_dir'],
        transport=__opts__['transport']
    )
    conn = get_conn()

    if 'instance_id' in vm_:
        # This was probably created via another process, and doesn't have
        # things like salt keys created yet, so let's create them now.
        if 'pub_key' not in vm_ and 'priv_key' not in vm_:
            log.debug('Generating minion keys for \'%s\'', vm_['name'])
            vm_['priv_key'], vm_['pub_key'] = __utils__['cloud.gen_keys'](
                config.get_cloud_config_value(
                    'keysize',
                    vm_,
                    __opts__
                )
            )
    else:
        # Put together all of the information required to request the instance,
        # and then fire off the request for it
        request_instance(conn=conn, call='action', vm_=vm_)
    data = show_instance(vm_.get('instance_id', vm_['name']), conn=conn, call='action')
    log.debug('VM is now running')

    def __query_node(vm_):
        data = show_instance(vm_['name'], conn=conn, call='action')
        if 'wait_for_metadata' in vm_:
            for key, value in six.iteritems(vm_.get('wait_for_metadata', {})):
                log.debug('Waiting for metadata: {0}={1}'.format(key, value))
                if data['metadata'].get(key, None) != value:
                    log.debug('Metadata is not ready: {0}={1}'.format(key, data['metadata'].get(key, None)))
                    return False
        return preferred_ip(vm_, data[ssh_interface(vm_)])
    try:
        ip_address = __utils__['cloud.wait_for_fun'](
            __query_node,
            vm_=vm_
        )
    except (SaltCloudExecutionTimeout, SaltCloudExecutionFailure) as exc:
        try:
            # It might be already up, let's destroy it!
            destroy(vm_['name'])
        except SaltCloudSystemExit:
            pass
        finally:
            raise SaltCloudSystemExit(six.text_type(exc))
    log.debug('Using IP address %s', ip_address)

    salt_interface = __utils__['cloud.get_salt_interface'](vm_, __opts__)
    salt_ip_address = preferred_ip(vm_, data[salt_interface])
    log.debug('Salt interface set to: %s', salt_ip_address)

    if not ip_address:
        raise SaltCloudSystemExit('A valid IP address was not found')

    vm_['ssh_host'] = ip_address
    vm_['salt_host'] = salt_ip_address

    ret = __utils__['cloud.bootstrap'](vm_, __opts__)
    ret.update(data)

    log.info('Created Cloud VM \'%s\'', vm_['name'])
    log.debug(
        '\'%s\' VM creation details:\n%s',
        vm_['name'], pprint.pformat(data)
    )

    event_data = {
        'name': vm_['name'],
        'profile': vm_['profile'],
        'provider': vm_['driver'],
        'instance_id': data['id'],
        'floating_ips': data['floating_ips'],
        'fixed_ips': data['fixed_ips'],
        'private_ips': data['private_ips'],
        'public_ips': data['public_ips'],
    }

    __utils__['cloud.fire_event'](
        'event',
        'created instance',
        'salt/cloud/{0}/created'.format(vm_['name']),
        args=__utils__['cloud.filter_event']('created', event_data, list(event_data)),
        sock_dir=__opts__['sock_dir'],
        transport=__opts__['transport']
    )
    __utils__['cloud.cachedir_index_add'](vm_['name'], vm_['profile'], 'nova', vm_['driver'])
    return ret


def destroy(name, conn=None, call=None):
    '''
    Delete a single VM
    '''
    if call == 'function':
        raise SaltCloudSystemExit(
            'The destroy action must be called with -d, --destroy, '
            '-a or --action.'
        )

    __utils__['cloud.fire_event'](
        'event',
        'destroying instance',
        'salt/cloud/{0}/destroying'.format(name),
        args={'name': name},
        sock_dir=__opts__['sock_dir'],
        transport=__opts__['transport']
    )

    if not conn:
        conn = get_conn()
    node = show_instance(name, conn=conn, call='action')
    log.info('Destroying VM: %s', name)
    ret = conn.delete_server(name)
    if ret:
        log.info('Destroyed VM: %s', name)
        # Fire destroy action
        __utils__['cloud.fire_event'](
            'event',
            'destroyed instance',
            'salt/cloud/{0}/destroyed'.format(name),
            args={'name': name},
            sock_dir=__opts__['sock_dir'],
            transport=__opts__['transport']
        )
        if __opts__.get('delete_sshkeys', False) is True:
            __utils__['cloud.remove_sshkey'](getattr(node, __opts__.get('ssh_interface', 'public_ips'))[0])
        if __opts__.get('update_cachedir', False) is True:
            __utils__['cloud.delete_minion_cachedir'](name, __active_provider_name__.split(':')[0], __opts__)
        __utils__['cloud.cachedir_index_del'](name)
        return True

    log.error('Failed to Destroy VM: %s', name)
    return False


def call(conn=None, call=None, kwargs=None):
    '''
    Call function from shade.

    func

        function to call from shade.openstackcloud library

    CLI Example

    .. code-block:: bash

        salt-cloud -f call myopenstack func=list_images
        t sujksalt-cloud -f call myopenstack func=create_network name=mysubnet
    '''
    if call == 'action':
        raise SaltCloudSystemExit(
            'The call function must be called with '
            '-f or --function.'
        )

    if 'func' not in kwargs:
        raise SaltCloudSystemExit(
            'No `func` argument passed'
        )

    if conn is None:
        conn = get_conn()

    func = kwargs.pop('func')
    for key, value in kwargs.items():
        try:
            kwargs[key] = salt.utils.json.loads(value)
        except ValueError:
            continue
    try:
        return getattr(conn, func)(**kwargs)
    except shade.exc.OpenStackCloudException as exc:
        log.error('Error running %s: %s', func, exc)
        raise SaltCloudSystemExit(six.text_type(exc))



# -*- coding: utf-8 -*-
'''
Manage CloudTrail Objects
=========================

.. versionadded:: 2016.11.0

Create and destroy CloudWatch event rules. Be aware that this interacts with Amazon's services,
and so may incur charges.

This module uses ``boto3``, which can be installed via package, or pip.

This module accepts explicit vpc credentials but can also utilize
IAM roles assigned to the instance through Instance Profiles. Dynamic
credentials are then automatically obtained from AWS API and no further
configuration is necessary. More information available `here
<http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html>`_.

If IAM roles are not used you need to specify them either in a pillar file or
in the minion's config file:

.. code-block:: yaml

    cloudwatch_event.keyid: GKTADJGHEIQSXMKKRBJ08H
    cloudwatch_event.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

It's also possible to specify ``key``, ``keyid`` and ``region`` via a profile,
either passed in as a dict, or as a string to pull from pillars or minion
config:

.. code-block:: yaml

    myprofile:
        keyid: GKTADJGHEIQSXMKKRBJ08H
        key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

.. code-block:: yaml

    Ensure event rule exists:
        boto_cloudwatch_event.present:
            - Name: mytrail
            - ScheduleExpression: 'rate(120 minutes)'
            - State: 'DISABLED'
            - Targets:
              - Id: "target1"
                Arn: "arn:aws:lambda:us-west-1:124456715622:function:my_function"
                Input: '{"arbitrary": "json"}'
            - region: us-east-1
            - keyid: GKTADJGHEIQSXMKKRBJ08H
            - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

'''

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals
import logging
import os

# Import Salt libs
import salt.utils.json

# Import 3rd-party libs
from salt.ext import six

log = logging.getLogger(__name__)


def __virtual__():
    '''
    Only load if boto is available.
    '''
    return 'boto_cloudwatch_event' if 'boto_cloudwatch_event.exists' in __salt__ else False


def present(name, Name=None,
           ScheduleExpression=None,
           EventPattern=None,
           Description=None,
           RoleArn=None,
           State=None,
           Targets=None,
           region=None, key=None, keyid=None, profile=None):
    '''
    Ensure trail exists.

    name
        The name of the state definition

    Name
        Name of the event rule.  Defaults to the value of the 'name' param if
        not provided.

    ScheduleExpression
        The scheduling expression. For example, ``cron(0 20 * * ? *)``,
        "rate(5 minutes)"

    EventPattern
        The event pattern.

    Description
        A description of the rule

    State
        Indicates whether the rule is ENABLED or DISABLED.

    RoleArn
        The Amazon Resource Name (ARN) of the IAM role associated with the
        rule.

    Targets
        A list of rresources to be invoked when the rule is triggered.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.
    '''
    ret = {'name': Name,
           'result': True,
           'comment': '',
           'changes': {}
           }

    Name = Name if Name else name

    if isinstance(Targets, six.string_types):
        Targets = salt.utils.json.loads(Targets)
    if Targets is None:
        Targets = []

    r = __salt__['boto_cloudwatch_event.exists'](Name=Name,
           region=region, key=key, keyid=keyid, profile=profile)

    if 'error' in r:
        ret['result'] = False
        ret['comment'] = 'Failed to create event rule: {0}.'.format(r['error']['message'])
        return ret

    if not r.get('exists'):
        if __opts__['test']:
            ret['comment'] = 'CloudWatch event rule {0} is set to be created.'.format(Name)
            ret['result'] = None
            return ret
        r = __salt__['boto_cloudwatch_event.create_or_update'](Name=Name,
                   ScheduleExpression=ScheduleExpression,
                   EventPattern=EventPattern,
                   Description=Description,
                   RoleArn=RoleArn,
                   State=State,
                   region=region, key=key, keyid=keyid, profile=profile)
        if not r.get('created'):
            ret['result'] = False
            ret['comment'] = 'Failed to create event rule: {0}.'.format(r['error']['message'])
            return ret
        _describe = __salt__['boto_cloudwatch_event.describe'](Name,
                                   region=region, key=key, keyid=keyid, profile=profile)
        if 'error' in _describe:
            ret['result'] = False
            ret['comment'] = 'Failed to create event rule: {0}.'.format(_describe['error']['message'])
            ret['changes'] = {}
            return ret
        ret['changes']['old'] = {'rule': None}
        ret['changes']['new'] = _describe
        ret['comment'] = 'CloudTrail {0} created.'.format(Name)

        if bool(Targets):
            r = __salt__['boto_cloudwatch_event.put_targets'](Rule=Name,
                   Targets=Targets,
                   region=region, key=key, keyid=keyid, profile=profile)
            if 'error' in r:
                ret['result'] = False
                ret['comment'] = 'Failed to create event rule: {0}.'.format(r['error']['message'])
                ret['changes'] = {}
                return ret
            ret['changes']['new']['rule']['Targets'] = Targets
        return ret

    ret['comment'] = os.linesep.join([ret['comment'], 'CloudWatch event rule {0} is present.'.format(Name)])
    ret['changes'] = {}
    # trail exists, ensure config matches
    _describe = __salt__['boto_cloudwatch_event.describe'](Name=Name,
                                  region=region, key=key, keyid=keyid, profile=profile)
    if 'error' in _describe:
        ret['result'] = False
        ret['comment'] = 'Failed to update event rule: {0}.'.format(_describe['error']['message'])
        ret['changes'] = {}
        return ret
    _describe = _describe.get('rule')

    r = __salt__['boto_cloudwatch_event.list_targets'](Rule=Name,
                   region=region, key=key, keyid=keyid, profile=profile)
    if 'error' in r:
        ret['result'] = False
        ret['comment'] = 'Failed to update event rule: {0}.'.format(r['error']['message'])
        ret['changes'] = {}
        return ret
    _describe['Targets'] = r.get('targets', [])

    need_update = False
    rule_vars = {'ScheduleExpression': 'ScheduleExpression',
                 'EventPattern': 'EventPattern',
                 'Description': 'Description',
                 'RoleArn': 'RoleArn',
                 'State': 'State',
                 'Targets': 'Targets'}
    for invar, outvar in six.iteritems(rule_vars):
        if _describe[outvar] != locals()[invar]:
            need_update = True
            ret['changes'].setdefault('new', {})[invar] = locals()[invar]
            ret['changes'].setdefault('old', {})[invar] = _describe[outvar]

    if need_update:
        if __opts__['test']:
            msg = 'CloudWatch event rule {0} set to be modified.'.format(Name)
            ret['comment'] = msg
            ret['result'] = None
            return ret

        ret['comment'] = os.linesep.join([ret['comment'], 'CloudWatch event rule to be modified'])
        r = __salt__['boto_cloudwatch_event.create_or_update'](Name=Name,
                   ScheduleExpression=ScheduleExpression,
                   EventPattern=EventPattern,
                   Description=Description,
                   RoleArn=RoleArn,
                   State=State,
                   region=region, key=key, keyid=keyid, profile=profile)
        if not r.get('created'):
            ret['result'] = False
            ret['comment'] = 'Failed to update event rule: {0}.'.format(r['error']['message'])
            ret['changes'] = {}
            return ret

        if _describe['Targets'] != Targets:
            removes = [i.get('Id') for i in _describe['Targets']]
            log.error(Targets)
            if bool(Targets):
                for target in Targets:
                    tid = target.get('Id', None)
                    if tid is not None and tid in removes:
                        ix = removes.index(tid)
                        removes.pop(ix)
                r = __salt__['boto_cloudwatch_event.put_targets'](Rule=Name,
                   Targets=Targets,
                   region=region, key=key, keyid=keyid, profile=profile)
                if 'error' in r:
                    ret['result'] = False
                    ret['comment'] = 'Failed to update event rule: {0}.'.format(r['error']['message'])
                    ret['changes'] = {}
                    return ret
            if bool(removes):
                r = __salt__['boto_cloudwatch_event.remove_targets'](Rule=Name,
                           Ids=removes,
                           region=region, key=key, keyid=keyid, profile=profile)
                if 'error' in r:
                    ret['result'] = False
                    ret['comment'] = 'Failed to update event rule: {0}.'.format(r['error']['message'])
                    ret['changes'] = {}
                    return ret
    return ret


def absent(name, Name=None, region=None, key=None, keyid=None, profile=None):
    '''
    Ensure CloudWatch event rule with passed properties is absent.

    name
        The name of the state definition.

    Name
        Name of the event rule.  Defaults to the value of the 'name' param if
        not provided.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.
    '''

    ret = {'name': Name,
           'result': True,
           'comment': '',
           'changes': {}
           }

    Name = Name if Name else name

    r = __salt__['boto_cloudwatch_event.exists'](Name,
                       region=region, key=key, keyid=keyid, profile=profile)
    if 'error' in r:
        ret['result'] = False
        ret['comment'] = 'Failed to delete event rule: {0}.'.format(r['error']['message'])
        return ret

    if r and not r['exists']:
        ret['comment'] = 'CloudWatch event rule {0} does not exist.'.format(Name)
        return ret

    if __opts__['test']:
        ret['comment'] = 'CloudWatch event rule {0} is set to be removed.'.format(Name)
        ret['result'] = None
        return ret

    # must remove all targets first
    r = __salt__['boto_cloudwatch_event.list_targets'](Rule=Name,
                                    region=region, key=key,
                                    keyid=keyid, profile=profile)
    if not r.get('targets'):
        ret['result'] = False
        ret['comment'] = 'Failed to delete event rule: {0}.'.format(r['error']['message'])
        return ret
    ids = [t.get('Id') for t in r['targets']]
    if bool(ids):
        r = __salt__['boto_cloudwatch_event.remove_targets'](Rule=Name, Ids=ids,
                                    region=region, key=key,
                                    keyid=keyid, profile=profile)
        if 'error' in r:
            ret['result'] = False
            ret['comment'] = 'Failed to delete event rule: {0}.'.format(r['error']['message'])
            return ret
        if r.get('failures'):
            ret['result'] = False
            ret['comment'] = 'Failed to delete event rule: {0}.'.format(r['failures'])
            return ret

    r = __salt__['boto_cloudwatch_event.delete'](Name,
                                    region=region, key=key,
                                    keyid=keyid, profile=profile)
    if not r['deleted']:
        ret['result'] = False
        ret['comment'] = 'Failed to delete event rule: {0}.'.format(r['error']['message'])
        return ret
    ret['changes']['old'] = {'rule': Name}
    ret['changes']['new'] = {'rule': None}
    ret['comment'] = 'CloudWatch event rule {0} deleted.'.format(Name)
    return ret



# -*- coding: utf-8 -*-
'''
Management of OpenStack Neutron Security Groups
===============================================

.. versionadded:: 2018.3.0

:depends: shade
:configuration: see :py:mod:`salt.modules.neutronng` for setup instructions

Example States

.. code-block:: yaml

    create security group;
      neutron_secgroup.present:
        - name: security_group1
        - description: "Very Secure Security Group"

    delete security group:
      neutron_secgroup.absent:
        - name_or_id: security_group1
        - project_name: Project1

    create security group with optional params:
      neutron_secgroup.present:
        - name: security_group1
        - description: "Very Secure Security Group"
        - project_id: 1dcac318a83b4610b7a7f7ba01465548

    create security group with optional params:
      neutron_secgroup.present:
        - name: security_group1
        - description: "Very Secure Security Group"
        - project_name: Project1
'''

from __future__ import absolute_import, print_function, unicode_literals

__virtualname__ = 'neutron_secgroup'


def __virtual__():
    if 'neutronng.list_subnets' in __salt__:
        return __virtualname__
    return (False, 'The neutronng execution module failed to load:\
                    shade python module is not available')


def present(name, auth=None, **kwargs):
    '''
    Ensure a security group exists.

    You can supply either project_name or project_id.

    Creating a default security group will not show up as a change;
    it gets created through the lookup process.

    name
        Name of the security group

    description
        Description of the security group

    project_name
        Name of Project

    project_id
        ID of Project

    '''
    ret = {'name': name,
           'changes': {},
           'result': True,
           'comment': ''}

    __salt__['neutronng.setup_clouds'](auth)

    if 'project_name' in kwargs:
        kwargs['project_id'] = kwargs['project_name']
        del kwargs['project_name']

    project = __salt__['keystoneng.project_get'](
        name=kwargs['project_id'])

    if project is None:
        ret['result'] = False
        ret['comment'] = "project does not exist"
        return ret

    secgroup = __salt__['neutronng.security_group_get'](
        name=name, filters={'tenant_id': project.id})

    if secgroup is None:
        if __opts__['test'] is True:
            ret['result'] = None
            ret['changes'] = kwargs
            ret['pchanges'] = ret['changes']
            ret['comment'] = 'Security Group will be created.'
            return ret

        secgroup = __salt__['neutronng.security_group_create'](**kwargs)
        ret['changes'] = secgroup
        ret['comment'] = 'Created security group'
        return ret

    changes = __salt__['neutronng.compare_changes'](secgroup, **kwargs)
    if changes:
        if __opts__['test'] is True:
            ret['result'] = None
            ret['changes'] = changes
            ret['pchanges'] = ret['changes']
            ret['comment'] = 'Security Group will be updated.'
            return ret

        __salt__['neutronng.security_group_update'](secgroup=secgroup, **changes)
        ret['changes'].update(changes)
        ret['comment'] = 'Updated security group'

    return ret


def absent(name, auth=None, **kwargs):
    '''
    Ensure a security group does not exist

    name
        Name of the security group

    '''
    ret = {'name': name,
           'changes': {},
           'result': True,
           'comment': ''}

    __salt__['neutronng.setup_clouds'](auth)

    kwargs['project_id'] = __salt__['keystoneng.project_get'](
        name=kwargs['project_name'])

    secgroup = __salt__['neutronng.security_group_get'](
        name=name,
        filters={'project_id': kwargs['project_id']}
    )

    if secgroup:
        if __opts__['test'] is True:
            ret['result'] = None
            ret['changes'] = {'id': secgroup.id}
            ret['pchanges'] = ret['changes']
            ret['comment'] = 'Security group will be deleted.'
            return ret

        __salt__['neutronng.security_group_delete'](name=secgroup)
        ret['changes']['id'] = name
        ret['comment'] = 'Deleted security group'

    return ret



# -*- coding: utf-8 -*-
'''
Manage Windows features via the ServerManager powershell module. Can list
available and installed roles/features. Can install and remove roles/features.

:maintainer:    Shane Lee <slee@saltstack.com>
:platform:      Windows Server 2008R2 or greater
:depends:       PowerShell module ``ServerManager``
'''

# Import Python libs
from __future__ import absolute_import, unicode_literals, print_function
import logging

try:
    from shlex import quote as _cmd_quote  # pylint: disable=E0611
except ImportError:
    from pipes import quote as _cmd_quote

# Import Salt libs
import salt.utils.json
import salt.utils.platform
import salt.utils.powershell
import salt.utils.versions
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)

__virtualname__ = 'win_servermanager'


def __virtual__():
    '''
    Load only on windows with servermanager module
    '''
    if not salt.utils.platform.is_windows():
        return False

    if salt.utils.versions.version_cmp(__grains__['osversion'], '6.1.7600') == -1:
        return False, 'Failed to load win_servermanager module: ' \
                      'Requires Remote Server Administration Tools which ' \
                      'is only available on Windows 2008 R2 and later.'

    if not salt.utils.powershell.module_exists('ServerManager'):
        return False, 'Failed to load win_servermanager module: ' \
                      'ServerManager module not available. ' \
                      'May need to install Remote Server Administration Tools.'

    return __virtualname__


def _pshell_json(cmd, cwd=None):
    '''
    Execute the desired powershell command and ensure that it returns data
    in JSON format and load that into python
    '''
    cmd = 'Import-Module ServerManager; {0}'.format(cmd)
    if 'convertto-json' not in cmd.lower():
        cmd = '{0} | ConvertTo-Json'.format(cmd)
    log.debug('PowerShell: %s', cmd)
    ret = __salt__['cmd.run_all'](cmd, shell='powershell', cwd=cwd)

    if 'pid' in ret:
        del ret['pid']

    if ret.get('stderr', ''):
        error = ret['stderr'].splitlines()[0]
        raise CommandExecutionError(error, info=ret)

    if 'retcode' not in ret or ret['retcode'] != 0:
        # run_all logs an error to log.error, fail hard back to the user
        raise CommandExecutionError(
            'Issue executing PowerShell {0}'.format(cmd), info=ret)

    # Sometimes Powershell returns an empty string, which isn't valid JSON
    if ret['stdout'] == '':
        ret['stdout'] = '{}'

    try:
        ret = salt.utils.json.loads(ret['stdout'], strict=False)
    except ValueError:
        raise CommandExecutionError(
            'No JSON results from PowerShell', info=ret)
    return ret


def list_available():
    '''
    List available features to install

    Returns:
        str: A list of available features as returned by the
        ``Get-WindowsFeature`` PowerShell command

    CLI Example:

    .. code-block:: bash

        salt '*' win_servermanager.list_available
    '''
    cmd = 'Import-Module ServerManager; ' \
          'Get-WindowsFeature ' \
          '-ErrorAction SilentlyContinue ' \
          '-WarningAction SilentlyContinue'
    return __salt__['cmd.shell'](cmd, shell='powershell')


def list_installed():
    '''
    List installed features. Supported on Windows Server 2008 and Windows 8 and
    newer.

    Returns:
        dict: A dictionary of installed features

    CLI Example:

    .. code-block:: bash

        salt '*' win_servermanager.list_installed
    '''
    cmd = 'Get-WindowsFeature ' \
          '-ErrorAction SilentlyContinue ' \
          '-WarningAction SilentlyContinue ' \
          '| Select DisplayName,Name,Installed'
    features = _pshell_json(cmd)

    ret = {}
    for entry in features:
        if entry['Installed']:
            ret[entry['Name']] = entry['DisplayName']

    return ret


def install(feature, recurse=False, restart=False, source=None, exclude=None):
    r'''
    Install a feature

    .. note::
        Some features require reboot after un/installation, if so until the
        server is restarted other features can not be installed!

    .. note::
        Some features take a long time to complete un/installation, set -t with
        a long timeout

    Args:

        feature (str, list):
            The name of the feature(s) to install. This can be a single feature,
            a string of features in a comma delimited list (no spaces), or a
            list of features.

            .. versionadded:: 2018.3.0
                Added the ability to pass a list of features to be installed.

        recurse (Options[bool]):
            Install all sub-features. Default is False

        restart (Optional[bool]):
            Restarts the computer when installation is complete, if required by
            the role/feature installed. Will also trigger a reboot if an item
            in ``exclude`` requires a reboot to be properly removed. Default is
            False

        source (Optional[str]):
            Path to the source files if missing from the target system. None
            means that the system will use windows update services to find the
            required files. Default is None

        exclude (Optional[str]):
            The name of the feature to exclude when installing the named
            feature. This can be a single feature, a string of features in a
            comma-delimited list (no spaces), or a list of features.

            .. warning::
                As there is no exclude option for the ``Add-WindowsFeature``
                or ``Install-WindowsFeature`` PowerShell commands the features
                named in ``exclude`` will be installed with other sub-features
                and will then be removed. **If the feature named in ``exclude``
                is not a sub-feature of one of the installed items it will still
                be removed.**

    Returns:
        dict: A dictionary containing the results of the install

    CLI Example:

    .. code-block:: bash

        # Install the Telnet Client passing a single string
        salt '*' win_servermanager.install Telnet-Client

        # Install the TFTP Client and the SNMP Service passing a comma-delimited
        # string. Install all sub-features
        salt '*' win_servermanager.install TFTP-Client,SNMP-Service recurse=True

        # Install the TFTP Client from d:\side-by-side
        salt '*' win_servermanager.install TFTP-Client source=d:\\side-by-side

        # Install the XPS Viewer, SNMP Service, and Remote Access passing a
        # list. Install all sub-features, but exclude the Web Server
        salt '*' win_servermanager.install "['XPS-Viewer', 'SNMP-Service', 'RemoteAccess']" True recurse=True exclude="Web-Server"
    '''
    # If it is a list of features, make it a comma delimited string
    if isinstance(feature, list):
        feature = ','.join(feature)

    # Use Install-WindowsFeature on Windows 2012 (osversion 6.2) and later
    # minions. Default to Add-WindowsFeature for earlier releases of Windows.
    # The newer command makes management tools optional so add them for parity
    # with old behavior.
    command = 'Add-WindowsFeature'
    management_tools = ''
    if salt.utils.versions.version_cmp(__grains__['osversion'], '6.2') >= 0:
        command = 'Install-WindowsFeature'
        management_tools = '-IncludeManagementTools'

    cmd = '{0} -Name {1} {2} {3} {4} ' \
          '-WarningAction SilentlyContinue'\
          .format(command, _cmd_quote(feature), management_tools,
                  '-IncludeAllSubFeature' if recurse else '',
                  '' if source is None else '-Source {0}'.format(source))
    out = _pshell_json(cmd)

    # Uninstall items in the exclude list
    # The Install-WindowsFeature command doesn't have the concept of an exclude
    # list. So you install first, then remove
    if exclude is not None:
        removed = remove(exclude)

    # Results are stored in a list of dictionaries in `FeatureResult`
    if out['FeatureResult']:
        ret = {'ExitCode': out['ExitCode'],
               'RestartNeeded': False,
               'Restarted': False,
               'Features': {},
               'Success': out['Success']}

        # FeatureResult is a list of dicts, so each item is a dict
        for item in out['FeatureResult']:
            ret['Features'][item['Name']] = {
                'DisplayName': item['DisplayName'],
                'Message': item['Message'],
                'RestartNeeded': item['RestartNeeded'],
                'SkipReason': item['SkipReason'],
                'Success': item['Success']
            }

            if item['RestartNeeded']:
                ret['RestartNeeded'] = True

        # Only items that installed are in the list of dictionaries
        # Add 'Already installed' for features that aren't in the list of dicts
        for item in feature.split(','):
            if item not in ret['Features']:
                ret['Features'][item] = {'Message': 'Already installed'}

        # Some items in the exclude list were removed after installation
        # Show what was done, update the dict
        if exclude is not None:
            # Features is a dict, so it only iterates over the keys
            for item in removed['Features']:
                if item in ret['Features']:
                    ret['Features'][item] = {
                        'Message': 'Removed after installation (exclude)',
                        'DisplayName': removed['Features'][item]['DisplayName'],
                        'RestartNeeded': removed['Features'][item]['RestartNeeded'],
                        'SkipReason': removed['Features'][item]['SkipReason'],
                        'Success': removed['Features'][item]['Success']
                    }

                    # Exclude items might need a restart
                    if removed['Features'][item]['RestartNeeded']:
                        ret['RestartNeeded'] = True

        # Restart here if needed
        if restart:
            if ret['RestartNeeded']:
                if __salt__['system.restart'](in_seconds=True):
                    ret['Restarted'] = True

        return ret

    else:

        # If we get here then all features were already installed
        ret = {'ExitCode': out['ExitCode'],
               'Features': {},
               'RestartNeeded': False,
               'Restarted': False,
               'Success': out['Success']}

        for item in feature.split(','):
            ret['Features'][item] = {'Message': 'Already installed'}

        return ret


def remove(feature, remove_payload=False, restart=False):
    r'''
    Remove an installed feature

    .. note::
        Some features require a reboot after installation/uninstallation. If
        one of these features are modified, then other features cannot be
        installed until the server is restarted. Additionally, some features
        take a while to complete installation/uninstallation, so it is a good
        idea to use the ``-t`` option to set a longer timeout.

    Args:

        feature (str, list):
            The name of the feature(s) to remove. This can be a single feature,
            a string of features in a comma delimited list (no spaces), or a
            list of features.

            .. versionadded:: 2018.3.0
                Added the ability to pass a list of features to be removed.

        remove_payload (Optional[bool]):
            True will cause the feature to be removed from the side-by-side
            store (``%SystemDrive%:\Windows\WinSxS``). Default is False

        restart (Optional[bool]):
            Restarts the computer when uninstall is complete, if required by the
            role/feature removed. Default is False

    Returns:
        dict: A dictionary containing the results of the uninstall

    CLI Example:

    .. code-block:: bash

        salt -t 600 '*' win_servermanager.remove Telnet-Client
    '''
    # If it is a list of features, make it a comma delimited string
    if isinstance(feature, list):
        feature = ','.join(feature)

    # Use Uninstall-WindowsFeature on Windows 2012 (osversion 6.2) and later
    # minions. Default to Remove-WindowsFeature for earlier releases of Windows.
    # The newer command makes management tools optional so add them for parity
    # with old behavior.
    command = 'Remove-WindowsFeature'
    management_tools = ''
    _remove_payload = ''
    if salt.utils.versions.version_cmp(__grains__['osversion'], '6.2') >= 0:
        command = 'Uninstall-WindowsFeature'
        management_tools = '-IncludeManagementTools'

        # Only available with the `Uninstall-WindowsFeature` command
        if remove_payload:
            _remove_payload = '-Remove'

    cmd = '{0} -Name {1} {2} {3} {4} ' \
          '-WarningAction SilentlyContinue'\
          .format(command, _cmd_quote(feature), management_tools,
                  _remove_payload,
                  '-Restart' if restart else '')
    try:
        out = _pshell_json(cmd)
    except CommandExecutionError as exc:
        if 'ArgumentNotValid' in exc.message:
            raise CommandExecutionError('Invalid Feature Name', info=exc.info)
        raise

    # Results are stored in a list of dictionaries in `FeatureResult`
    if out['FeatureResult']:
        ret = {'ExitCode': out['ExitCode'],
               'RestartNeeded': False,
               'Restarted': False,
               'Features': {},
               'Success': out['Success']}

        for item in out['FeatureResult']:
            ret['Features'][item['Name']] = {
                'DisplayName': item['DisplayName'],
                'Message': item['Message'],
                'RestartNeeded': item['RestartNeeded'],
                'SkipReason': item['SkipReason'],
                'Success': item['Success']
            }

        # Only items that installed are in the list of dictionaries
        # Add 'Not installed' for features that aren't in the list of dicts
        for item in feature.split(','):
            if item not in ret['Features']:
                ret['Features'][item] = {'Message': 'Not installed'}

        return ret

    else:

        # If we get here then none of the features were installed
        ret = {'ExitCode': out['ExitCode'],
               'Features': {},
               'RestartNeeded': False,
               'Restarted': False,
               'Success': out['Success']}

        for item in feature.split(','):
            ret['Features'][item] = {'Message': 'Not installed'}

        return ret
